/******************************************************************************/
/*
 *
 * @file fip_urv.c
 *
 * Hydra software
 *
 ******************************************************************************/

#include <stdint.h>
#include "common.h"

/* Produced variable length (the reply) */
#define VAR_LENGTH      16
#define CONS_VAR_LENGTH 16

/* Magic number for applications.  */
#define APP_MAGIC 0xb1507d38

/* Total bytes read from I2C */
#define SB_TOTAL_BYTES  (ADC_SB_I2C_BYTES \
                        + FANTRAY_I2C_BYTES \
                        + RAD7291_I2C_BYTES)

/* Align system buffer size */
#define SB_BUFFER_BYTES ((SB_TOTAL_BYTES + 3) & ~3)

/**************************** Type Definitions ********************************/
struct i2c_cmd {
    unsigned char state;                    //I2C current state
    unsigned char bus;
    unsigned char addr;
    unsigned char len;                      //I2C length
    unsigned char coff;                     //I2C command offset
    unsigned char boff;                     //I2C buffer offset
    unsigned char buf[SB_BUFFER_BYTES];     //I2C buffer
    const unsigned char *cmds;              //Command list
} __attribute__((aligned(4)));

/************************** Variable Definitions ******************************/

/* Data collected during i2c scan */
static unsigned char system_buf[SB_BUFFER_BYTES+4];

/* Set when the host is reading the system_buf buffer. The buffer cannot
   be written when this flag is set so that the host has a coherent set of
   values. */
static unsigned char system_read_flag;

/*  I2C scan status.  It's a state machine.  Transitions are:
    DISABLE -> ENABLE : enable scanning (by a user user)
    ENABLE -> DISABLE : disable scanning (by a user command)
    ENABLE -> RUN     : scan (on a timeout)
    RUN -> ENABLE     : scan done
    RUN -> RESET      : on i2c error (power-off i2c bus)
    RESET -> ENABLE   : after error
*/
#define I2C_SCAN_DISABLE    0x0    /* never (allowed for user) */
#define I2C_SCAN_ENABLE     0x1    /* scan (forbidden to user) */
#define I2C_SCAN_RUN        0x2    /* scanning */
#define I2C_SCAN_RESET      0x3    /* in reset */
static unsigned char i2c_bus_status;

/* i2c devices to scan */
static unsigned i2c_device_enable;

/* Mask of devices to be scanned */
static unsigned i2c_device_scan;

/* Devices to be reset */
static unsigned i2c_device_reset;


/* Define Array with Statistics */
static unsigned stat_regs[STAT_NBR];

/*
 *  i2c commands to initialize adc in the system board
 */
static const unsigned char initialize_adc_sb_cmd[] = {
    I2C_CMD_SET_TARG | (ADC_SB_I2C_BUS_ID << 4), ADC_SB_I2C_ADDR,
    I2C_CMD_START_W,
    I2C_CMD_WRITE_STOP, 3, 0x00, 0xFF, 0xA0, /* Init reg 0x0 */

    I2C_CMD_IDLE
};

/*
 *  i2c commands to initialize ratopus
 */
static const unsigned char initialize_ratopus_cmd[] = {
    I2C_CMD_SET_TARG | (RAD7291_I2C_BUS_ID << 4), RAD7291_I2C_ADDR,
    I2C_CMD_START_W,
    I2C_CMD_WRITE_STOP, 3, 0x00, 0xF0, 0xA0, /* Init reg 0x0 */

    I2C_CMD_IDLE
};

/*
 *  i2c commands that are executed to get values from i2c devices
 *  -------------------------------------------------------------
 *      -[0x50]: R/W register to Enable/Disable the scanning of an i2c device.
 *      Default value is 0
 *          - bit 0 : enable scan system board ADC
 *          - bit 1 : enable scan Fantray
 *          - bit 2 : enable scan RaTopus ADC
 */

/*
 *  System board ADC (1)
 */
static const unsigned char i2c_adc_cmds[] = {

    /********************     SYSTEM BOARD ADC SEQUENCE     *******************/
    I2C_CMD_SET_TARG | (ADC_SB_I2C_BUS_ID << 4), ADC_SB_I2C_ADDR,

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x01,         /* Voltage Conversion ADC */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 16,          /* Read 16 bytes */

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x02,         /* Temperature */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,           /* Read 2 bytes */

    I2C_CMD_IDLE

};

/*
 *  Fantray (2)
 */
static const unsigned char i2c_ft_cmds[] = {

    /*********************     FANTRAY MONIMOD SEQUENCE     *******************/
    I2C_CMD_SET_TARG | (FANTRAY_I2C_BUS_ID << 4), FANTRAY_I2C_ADDR,

    I2C_CMD_START_W,
    I2C_CMD_WRITE_STOP, 2, 0x00, 0x00,  /* page 0 */

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x8b,             /* vout */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x8c,             /* iout */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,

    I2C_CMD_START_W,
    I2C_CMD_WRITE_STOP, 2, 0x00, 0x01,  /* page 1 */

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x8b,             /* vout */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x8d,             /* temp-1 */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x90,             /* Fan Speed 1 */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x91,             /* Fan Speed 2 */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x92,             /* Fan Speed 3 */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,

    I2C_CMD_IDLE

};

/*
 *  RaTopus
 */
static const unsigned char i2c_rtp_cmds[] = {

    /***********************     RATOPUS SEQUENCE     ************************/
    I2C_CMD_SET_TARG | (RAD7291_I2C_BUS_ID << 4), RAD7291_I2C_ADDR,

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x01,             /* Voltage Conversion Ratopus */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 8,               /* Read 8 bytes */

    I2C_CMD_START_W,
    I2C_CMD_WRITE, 1, 0x02,             /* Temperature */
    I2C_CMD_START_R,
    I2C_CMD_READ_STOP, 2,               /* Read 2 bytes */

    I2C_CMD_IDLE
};


/******************************* Functions ************************************/
/* Save regs to memory to detect any divergence */
extern void refresh_regs(void);

/*
 *  Function:    extract_word
 *  -------------------------
 *  Convert 4 bytes in 32 bits word
 *
 *  buf: pointer to the buffer
 *  res: return 32 bits word
 *
 */
static unsigned extract_word (volatile uint32_t *buf)
{
    unsigned res;

    /* Little endian */
    res = (buf[0] & 0xff) << 0;
    res |= (buf[1] & 0xff) << 8;
    res |= (buf[2] & 0xff) << 16;
    res |= (buf[3] & 0xff) << 24;

    return res;
}

/*
 *  Function:    insert_word
 *  ------------------------
 *  Convert 32 bits word to 4 bytes
 *
 *  buf: pointer to the buffer
 *  val: value in 32 bits
 *
 */
static void insert_word(unsigned char *buf, unsigned val)
{
    buf[0] = (val >> 0) & 0xff;
    buf[1] = (val >> 8) & 0xff;
    buf[2] = (val >> 16) & 0xff;
    buf[3] = (val >> 24) & 0xff;
}

/*
 *  Function:    do_i2c_write
 *  -------------------------
 *
 *
 *  i2c_cmd: struct with the actions which are done in the I2C bus
 *
 */
static inline void do_i2c_write (struct i2c_cmd *i2c)
{
    /* Set the stop bit only on the last byte and if requested */
    unsigned char stop = i2c->len == 1 && i2c->state == I2C_CMD_WRITE_STOP;

    REGS->i2c[OC_I2C_RXTX] = i2c->cmds[i2c->coff];
    REGS->i2c[OC_I2C_CR] = 0x10 | (stop ? 0x40 : 0);  /* WRITE */
    i2c->coff++;
}

/*
 *  Function:    do_i2c_read
 *  ------------------------
 *
 *
 *  i2c_cmd: struct with the actions which are done in the I2C bus
 *
 */
static inline void do_i2c_read (struct i2c_cmd *i2c)
{
    /* Set the stop bit only on the last byte and if requested */
    unsigned char stop = i2c->len == 1 && i2c->state != I2C_CMD_READ;

    REGS->i2c[OC_I2C_CR] = 0x20 | (stop ? 0x48 : 0);  /* READ */
}

/*
 *  Function:    do_i2c_done
 *  -----------------------
 *
 *
 *  i2c_cmd: struct with the actions which are done in the I2C bus
 *  stat: status of the IP Core
 *
 */
static void do_i2c_done (struct i2c_cmd *i2c, unsigned char stat)
{
    /*
     *  According to the current command (state) check if it has been
     *  executed, and the status
     */
    if (i2c->state == I2C_CMD_START_R || i2c->state == I2C_CMD_START_W) {

        if (stat == 0x41) {
            /* Busy + Interrupt: ok, can start; next command */
            i2c->state = I2C_CMD_NEXT;
            return;
        }

    } else if (i2c->state == I2C_CMD_WRITE
               || i2c->state == I2C_CMD_WRITE_STOP) {

        if (stat == 0x41 || (stat == 0x01 && i2c->len == 1)) {

            /* Next byte */
            i2c->len--;
            if (i2c->len == 0) {
                i2c->state = I2C_CMD_NEXT;
            } else {
                do_i2c_write(i2c);
            }

            return;

        }

    } else if (i2c->state == I2C_CMD_READ
               || i2c->state == I2C_CMD_READ_STOP
               || i2c->state == I2C_CMD_READLEN_STOP) {

        /*
         *  We are the master. No need to check STAT, as the ACK/BUSY bits are
         *  in fact set by us.
         *  Read the byte received.
         */
        unsigned v;

        /* Read the byte */
        v = REGS->i2c[OC_I2C_RXTX];
        i2c->buf[i2c->boff] = v;
        i2c->boff++;

        if (i2c->state == I2C_CMD_READLEN_STOP)
            i2c->len = v;
        else
            i2c->len--;

        if (i2c->len == 0) {
            /* All the bytes have been read, next i2c command */
            i2c->state = I2C_CMD_NEXT;
        } else {
            /* Prepare to read next byte */
            do_i2c_read(i2c);
        }

        return;
    }

    /* Bad stat value, it's an error */

    i2c->state = I2C_CMD_ERR;
    REGS->i2c[OC_I2C_CR] = 0x40; /* Stop */

    if (i2c_bus_status == I2C_SCAN_RUN) {

        /* This is an error during monimod scan, shutdown monimod */
        if (i2c->bus == ADC_SB_I2C_BUS_ID) {
            REGS->i2c_off |= FIP_URV_REGS_I2C_OFF_MONIMOD;
            REGS->monimod_shdn = FIP_URV_REGS_MONIMOD_SHDN_VAL;
            stat_regs[STAT_ADC_RST]++;
        }
        else if (i2c->bus == RAD7291_I2C_BUS_ID) {
            stat_regs[STAT_RATOPUS_ERR]++;
        }
        else if (i2c->bus == FANTRAY_I2C_BUS_ID) {
            REGS->i2c_off |= FIP_URV_REGS_I2C_OFF_FANTRAY;
            REGS->fan_ctrl = FIP_URV_REGS_FAN_CTRL_RSTN;
            stat_regs[STAT_FANTRAY_RST]++;
        }

        i2c_bus_status = I2C_SCAN_RESET;
    }
}

static void do_i2c_load_scan_cmd(struct i2c_cmd *i2c)
{
    if (i2c_device_enable & i2c_device_scan & I2C_SLV_ADC) {
        if (i2c_device_reset & I2C_SLV_ADC) {
            i2c->cmds = initialize_adc_sb_cmd;
        }
        else
            i2c->cmds = i2c_adc_cmds;
        i2c_device_scan &= ~I2C_SLV_ADC;
    }
    else if (i2c_device_enable & i2c_device_scan & I2C_SLV_FT) {
        i2c->cmds = i2c_ft_cmds;
        i2c_device_scan &= ~I2C_SLV_FT;
    }
    else if (i2c_device_enable & i2c_device_scan & I2C_SLV_RTP) {
        if (i2c_device_reset & I2C_SLV_RTP) {
            i2c->cmds = initialize_ratopus_cmd;
        }
        else
            i2c->cmds = i2c_rtp_cmds;
        i2c_device_scan &= ~I2C_SLV_RTP;
    }

    /* It's a new command, decode it from the beginning */
    i2c->coff = 0;
}

/*
 * Function:    do_i2c_run
 * -----------------------
 *  Scan automatically all the I2C devices: ADC, Fantray, Ratopus is they are
 *  defined
 *
 *  i2c_cmd: struct with the actions which are done in the I2C bus
 *
 */
static void do_i2c_run (struct i2c_cmd *i2c)
{
    /* Read the command to be executed */
    unsigned char cmd_i2c = i2c->cmds[i2c->coff];

    /* And skip it */
    i2c->coff++;

    if ((cmd_i2c & 0x0f) == I2C_CMD_SET_TARG) {
        /* Set target */
        i2c->bus = cmd_i2c >> 4;
        REGS->i2c[OC_I2C_CFG] = i2c->bus;

        /* This is a prefix command.  The next byte contains the address */
        cmd_i2c = I2C_CMD_SET_ADDR;
    }

    if (cmd_i2c == I2C_CMD_SET_ADDR) {
        /* Set address */
        i2c->addr = i2c->cmds[i2c->coff];
        i2c->coff++;

        /* This is a prefix command without any i2c transmission.  Do not
           wait for busy bit, execute now the next command. */
        cmd_i2c = i2c->cmds[i2c->coff];
    }

    /* Default state: the command */
    i2c->state = cmd_i2c;
    i2c->len = 0;

    if (cmd_i2c == I2C_CMD_IDLE) {
        /* Last command */

        if (i2c_bus_status != I2C_SCAN_RUN) {
            /* User command, stop now  */
            return;
        }

        /* Was monitor scanning.  */

        if (i2c_device_scan != 0) {
            /* Scan is not finished, next device */
            do_i2c_load_scan_cmd(i2c);

            /* Begin for the start.  */
            i2c->coff = 0;

            cmd_i2c = i2c->cmds[i2c->coff];
            i2c->state = cmd_i2c;

            /* Continue execution (hopefully tail-call) */
            do_i2c_run(i2c);
            return;
        }
        else {
            /* End of scan cycle.  Copy received data from i2c buffer
               to system buffer (so that data won't be modified while being
               sent) */
            unsigned i;

            if (!system_read_flag) {
                system_buf[0] = i2c->boff + 4;
                system_buf[1] = i2c_device_enable | (i2c_device_reset << 4);
                system_buf[2]++; /* Change cycle */

                /* Copy result, so that a consistent view can be extracted */
                for (i = 0; i < sizeof(i2c->buf); i++) {
                    system_buf[4 + i] = i2c->buf[i];
                }
            }

            i2c_bus_status = I2C_SCAN_ENABLE;   /* Leave that mode */

            /* Devices have been reset */
            i2c_device_reset = 0;

            return;
        }
    }

    if (cmd_i2c == I2C_CMD_START_R || cmd_i2c == I2C_CMD_START_W) {

        /* Start set of commands */

        /* Set bit0 if the command is a read. */
        REGS->i2c[OC_I2C_RXTX] = i2c->addr | (cmd_i2c == I2C_CMD_START_R);
        REGS->i2c[OC_I2C_CR] = 0x90;  /* START + WRITE */

    } else if (cmd_i2c == I2C_CMD_WRITE || cmd_i2c == I2C_CMD_WRITE_STOP) {

        /* Write Operation */
        i2c->len = i2c->cmds[i2c->coff];
        i2c->coff++;
        do_i2c_write(i2c);

    } else if (cmd_i2c == I2C_CMD_READ
               || cmd_i2c == I2C_CMD_READ_STOP
               || cmd_i2c == I2C_CMD_READLEN_STOP) {

        /* Read Operation */
        if (cmd_i2c != I2C_CMD_READLEN_STOP) {
            i2c->len = i2c->cmds[i2c->coff];
            i2c->coff++;
        }
        do_i2c_read(i2c);

    } else {

        /* Bad command */
        i2c->state = I2C_CMD_ERR;

    }
}

/*
 * Function:    copy_fip_to_app
 * ----------------------------
 *
 */
static void copy_fip_to_app(unsigned cmd)
{
    unsigned lmb = REGS->app[0].mailboxes.mboxout; /* written by mgmt */
    unsigned rmb = REGS->app[0].mailboxes.mboxin;  /* written by app */
    unsigned wslot = (lmb >> 8) & 0xff;
    unsigned rslot = rmb & 0xff;
    unsigned nslot = (wslot + 1) & 3;
    volatile uint32_t *dest;
    volatile uint32_t *src;
    unsigned i;
    unsigned d;

    if (nslot == rslot) {
        /* No free slot, discard. */
        stat_regs[STAT_DATA_WDIS]++;
        return;
    }

    /* Pack and copy.  */
    dest = &REGS->sh_mem[wslot << 5];
    d = cmd & 0x0f;
    src = &REGS->fip_reg[0x2];

    for (i = 0; i < CONS_VAR_LENGTH; i += 4) {
        if (i > 0) {
                src += 4;
                d = src[0];
        }

        d |= src[1] << 8;
        d |= src[2] << 16;
        d |= src[3] << 24;
        *dest++ = d;
    }

    /* Update slot.  */
    REGS->app[0].mailboxes.mboxout = (lmb + (1 << 8)) & 0xffff03ff;
    stat_regs[STAT_DATA_WR]++;
}

/*
 * Function:    copy_app_to_fip
 * ----------------------------
 *
 *
 */
static int copy_app_to_fip(void)
{
    unsigned rmb = REGS->app[0].mailboxes.mboxin;  /* written by app */
    unsigned lmb = REGS->app[0].mailboxes.mboxout; /* written by mgmt */
    unsigned wslot = (rmb >> 8) & 0xff;
    unsigned rslot = lmb & 0xff;
    volatile uint32_t *src;
    unsigned i;
    unsigned d;

    /* Maybe nothing to write. */
    if (rslot == wslot)
        return 0;

    src = &REGS->sh_mem[128 + (rslot << 5)];

    d = *src++;
    d &= 0xffffff0f;
    d |= CMD_APP_FLAG;

    /* Produce variable */
    REGS->fip_var3 = 1;

    /* Unpack and copy */
    i = 0;
    while (1) {
        REGS->fip_reg[0x102 + i] = d;
        i++;
        if (i == VAR_LENGTH)
            break;
        if ((i & 3) == 0)
            d = *src++;
        else
            d >>= 8;
    }

    REGS->fip_var3 = 0;

    /* Update slot.  */
    REGS->app[0].mailboxes.mboxout = (lmb + 1) & 0xffffff03;

    return 1;
}

/*
 * Function:    load_app_cpu
 * -------------------------
 *
 *
 */
static void load_app_cpu(void)
{
    uint32_t *envm = (uint32_t *)REGS->envm;
    unsigned off;
    unsigned i;
    unsigned len;

    /* Try to find the magic number.  */
    off = 0x4000 >> 2;
    if (envm[off + 1] != APP_MAGIC) {
        off = 0x6000 >> 2;
        if (envm[off + 1] != APP_MAGIC) {
            return;
        }
    }

    /* Start address.  */
    stat_regs[STAT_APP0_START] = off << 2;

    /* Copy.
       TODO: protection against errors ? */
    len = REGS->config_ram & FIP_URV_REGS_CONFIG_RAM_APP0_IRAM_MASK;
    len = 1 << (len + 8 - 2);
    for (i = 0; i < len; i++) {
        REGS->app0_mem[i] = envm[off + i];
        SUPERVISOR->wd_key = WD_KEY;
    }

    /* Start */
    REGS->app[0].cr = 0;
}

/********************************** MAIN **************************************/
int main (void)
{
    unsigned timeout_led;
    unsigned char reply[VAR_LENGTH];
    unsigned char rdy;
    unsigned char leds;
    unsigned char rlen;

    static struct i2c_cmd i2c;
    static unsigned timeout_i2c;
    static unsigned char user_i2c_cmds[USER_I2C_LEN];

#if VAR_LENGTH == 8
  /* That's the default */
#elif VAR_LENGTH == 16
  REGS->fip_var = 2;
  SUPERVISOR->wd_period = 5000;
#elif VAR_LENGTH == 64
  REGS->fip_var = 4;
  SUPERVISOR->wd_period = 5000;
#elif VAR_LENGTH == 124
  REGS->fip_var = 5;
  SUPERVISOR->wd_period = 8500;
#else
# error "Incorrect VAR_LENGTH"
#endif

    timeout_led = 0;
    rdy = 0;

    /* Var1 has been consumed so it's released in case of timeout */
    REGS->fip_var1 = 0;

    /* Clear recovery bit */
    SUPERVISOR->cpu = 0;

    if (SUPERVISOR->reset_cause == 0) {

        /* POWER UP */
        leds = 1;

        /* Initialize i2c Struct */
        i2c.state = I2C_CMD_IDLE;
        i2c.coff = 0;
        i2c.boff = 0;
        i2c.len = 0;
        i2c.cmds = user_i2c_cmds;

        /* Initialize Registers */
        timeout_i2c = 0;
        stat_regs[STAT_MIN_WD_CNT] = ~0;

        /* Scan disable before set up the register.
           TODO: enable at least adc by default ? */
        i2c_bus_status = I2C_SCAN_DISABLE;
        i2c_device_enable = 0;
        i2c_device_reset = I2C_SLV_ADC | I2C_SLV_RTP;

        /* Power-on monimod */
        REGS->monimod_shdn = 0;

        /* Power on fantray */
        REGS->fan_ctrl &= ~FIP_URV_REGS_FAN_CTRL_RSTN;

        /* Pull up I2C */
        REGS->i2c_off = 0;

        /* Enable scrubbers */
        SUPERVISOR->scrub_cfg = HYDRA_SUPERVISOR_REGS_SCRUB_CFG_IRAM_EN
                            | HYDRA_SUPERVISOR_REGS_SCRUB_CFG_DRAM_EN;

        REGS->shmem_regs.scrub_cfg = FIP_URV_REGS_SHMEM_REGS_SCRUB_CFG_EN;

        /* Init i2c: prescaler (400khz) */
        REGS->i2c[OC_I2C_PRES_LO] = 25;
        REGS->i2c[OC_I2C_PRES_HI] = 0;
        REGS->i2c[OC_I2C_CTRL] = 0x80;  /* Enable */

        /* Build date stat */
        stat_regs[STAT_BUILD_DATE] = C_BUILD_DATE;
        stat_regs[STAT_GIT_HASH] = C_GIT_HASH;

        /* Load application from eNVM. */
        load_app_cpu();
    } else {

        leds = 3;
        stat_regs[STAT_RESET]++;

    }

    /*************************     Infinite loop     **************************/
    while (1) {

        unsigned cpu;

        /* Check for FIP message */
        unsigned int fip_status;

        /* Get minimum watchdog count */
        unsigned count = SUPERVISOR->wd_count;
        if (count < stat_regs[STAT_MIN_WD_CNT]) {
            stat_regs[STAT_MIN_WD_CNT] = count;
        }

        SUPERVISOR->wd_key = WD_KEY;

        refresh_regs();

        stat_regs[STAT_CYCLE_CNT]++;

        /* Force recovery in case of divergence */
        cpu = SUPERVISOR->cpu;
        if (cpu != 0) {
            stat_regs[STAT_RECOVERY]++;
            stat_regs[STAT_LAST_RST] = cpu;
            asm volatile ("");
            SUPERVISOR->cpu = HYDRA_SUPERVISOR_REGS_CPU_RECOVERY;
        }

        fip_status = REGS->fip_status;

        /* Control leds */
        if (leds & 0x80) {
            REGS->leds = leds;
        } else {
            /* Leds 3 & 2 are for liveness */
            REGS->leds = leds;
            timeout_led++;
            if (timeout_led == 100000){
                leds = (leds & 1) ^ 1;
                timeout_led = 0;
            }
        }

        if (fip_status & 1) {

            /* Variable var1 received (from FIP) */
            unsigned char cmd;
            unsigned char id;

            /* Consume variable */
            REGS->fip_var1 = 1;
            cmd = REGS->fip_reg[0x02];

            if (cmd & CMD_APP_FLAG) {
                /* Just data.  */
                copy_fip_to_app (cmd);
            } else {

                id = REGS->fip_reg[0x03];

                /* Default reply: command + id */
                reply[0] = cmd;
                reply[1] = id;
                reply[3] = REGS->fip_reg[0x02 + CONS_VAR_LENGTH]; /* MPS */
                /* MPS: bit 0: refreshment, bit 2: significance */

                /* A reply is ready. */
                rdy = 1;

                /* Default reply length */
                rlen = 8;

                stat_regs[STAT_FIP_CMD]++;

                /* No switch (smaller)  */
                if (cmd == CMD_SET_LEDS) {

                    leds = REGS->fip_reg[0x06];

                } else if (cmd == CMD_READ_WORD) {

                    volatile unsigned *addr;
                    addr = (unsigned *)extract_word(&REGS->fip_reg[0x6]);
                    insert_word(&reply[4], *addr);

                } else if (cmd == CMD_WRITE_WORD) {

                    volatile unsigned *addr;
                    unsigned val;
                    unsigned k;
                    unsigned l;

                    stat_regs[STAT_FIP_WRITES]++;

                    l = REGS->fip_reg[0x02 + 2];
                    addr = (unsigned *)extract_word(&REGS->fip_reg[0x02 + 4]);

                    stat_regs[STAT_WRITE_ADDR] = (unsigned)addr;

                    for (k = 4; k < l; k += 4) {
                        val = extract_word(&REGS->fip_reg[0x02 + 0x04 + k]);
                        addr[(k - 4) >> 2] = val;
                    }

                } else if (cmd == CMD_WRITE_ECC) {

                    volatile unsigned *addr;
                    unsigned val;
                    unsigned ecc;

                    stat_regs[STAT_FIP_WRITES]++;
                    addr = (unsigned *)extract_word(&REGS->fip_reg[0x6]);
                    stat_regs[STAT_WRITE_ADDR] = (unsigned)addr | 1;

                    val = extract_word(&REGS->fip_reg[0x6 + 4]);
                    ecc = REGS->fip_reg[0x6 + 8];
                    SUPERVISOR->dram_ecc_mask = ecc;
                    *addr = val;
                    SUPERVISOR->dram_ecc_mask = 0;

                } else if (cmd == CMD_I2C_MON) {

                    unsigned val = REGS->fip_reg[0x06];
                    if (val < 2 && i2c_bus_status < I2C_SCAN_RUN) {
                        /* Update */
                        i2c_device_enable = REGS->fip_reg[0x07];
                        i2c_device_reset = REGS->fip_reg[0x08];

                        /* Can only disable scanning, when idle. */
                        i2c_bus_status = val;

                    }
                    reply[4] = i2c_bus_status;
                    reply[5] = i2c_device_enable;

                } else if (cmd == CMD_I2C_TFR) {

                    if (i2c_bus_status != I2C_SCAN_DISABLE
                        || i2c.state != I2C_CMD_IDLE) {

                        /* Busy.
                           (either in scan mode, or a command is running) */
                        reply[4] = 1;

                    } else {

                        /* Extract i2c commands from fip */
                        unsigned k;
                        unsigned clen = REGS->fip_reg[0x6];
                        for (k = 0; k < clen; k++) {
                            user_i2c_cmds[k] = REGS->fip_reg[0x7 + k];
                        }

                        user_i2c_cmds[clen] = I2C_CMD_IDLE;
                        i2c.coff = 0;
                        i2c.boff = 0;
                        i2c.cmds = user_i2c_cmds;

                        /* Has to execute i2c commands */
                        i2c.state = I2C_CMD_NEXT;
                        /* fip reply */
                        reply[4] = 0;

                    }

                } else if (cmd == CMD_I2C_STAT) {

                    reply[4] = i2c.state;
                    reply[5] = i2c.len;
                    reply[6] = i2c.coff;
                    reply[7] = i2c.boff;

                } else if (cmd == CMD_I2C_BUF) {

                    unsigned off = REGS->fip_reg[0x6];
                    reply[4] = i2c.buf[off + 0];
                    reply[5] = i2c.buf[off + 1];
                    reply[6] = i2c.buf[off + 2];
                    reply[7] = i2c.buf[off + 3];

                } else if (cmd == CMD_MON_BUF) {

                    unsigned off = REGS->fip_reg[0x6];
                    system_read_flag = REGS->fip_reg[0x7];

                    reply[4] = system_buf[off + 0];
                    reply[5] = system_buf[off + 1];
                    reply[6] = system_buf[off + 2];
                    reply[7] = system_buf[off + 3];

                } else if (cmd == CMD_STAT_BUF) {

                    unsigned off = REGS->fip_reg[0x6];
                    insert_word(&reply[4], stat_regs[off]);

                } else if (cmd == CMD_CPU_FLT) {

                    unsigned cpu = REGS->fip_reg[0x6];
                    if (SUPERVISOR->force_divergence != cpu) {
                        reply[4] = 1;
                    }

                } else if (cmd == CMD_FIP_PING) {

                    unsigned i;
                    rlen = VAR_LENGTH;
                    reply[2] = rlen;
                    for (i = 4; i < rlen; i++) {
                        reply[i] = ~REGS->fip_reg[0x02 + i];
                    }

                } else if (cmd == CMD_WAIT) {

                    unsigned val;
                    unsigned param;

                    val = extract_word(&REGS->fip_reg[0x6]);
                    param = extract_word(&REGS->fip_reg[0x6 + 4]);
                    SUPERVISOR->wd_key = WD_KEY;

                    while (val-- != 0) {
                        if (param & 1) {
                            SUPERVISOR->wd_key = WD_KEY;
                        }
                    }

                } else {

                    /* No reply for pad (or error).  */
                    reply[0] = CMD_ERROR;
                    rdy = 0;

                }
            }

            /* Variable has been consumed */
            REGS->fip_var1 = 0;

        }

        if (fip_status & 2) {

            /* var3 can be written. */
            /* High priority: answer packet from management. */
            if (rdy) {

                int i;

                /* Produce variable */
                REGS->fip_var3 = 1;

                /* Write a reply. */
                for (i = 0; i < rlen; i++) {
                    REGS->fip_reg[0x102 + i] = reply[i];
                }

                REGS->fip_var3 = 0;
                rdy = 0;

            } else if (copy_app_to_fip()) {

                /* done */

            } else {

                /* Nothing to do; do not send data, clear var3_rdy.  */
                REGS->fip_var3 = 2;

            }

        }

        /* i2c scanning */

        if ( (i2c.state != I2C_CMD_IDLE) && (i2c.state != I2C_CMD_ERR) ) {
            /* An i2c command is executed */

            if (i2c.state != I2C_CMD_NEXT) {

                /* For end of transfer, check TIP (transfer in progess)  */
                unsigned char stat = REGS->i2c[OC_I2C_CR];
                if (!(stat & 2)) {
                    do_i2c_done (&i2c, stat);
                }

            }

            /*
             *  Execute the command. As do_i2c_done can change the state,
             *  the state needs to be checked again
             */
            if (i2c.state == I2C_CMD_NEXT) {
                do_i2c_run (&i2c);
            }

        }
        else if (i2c_bus_status == I2C_SCAN_ENABLE
                 && i2c_device_enable != 0
                 && timeout_i2c++ == 200000) {

            /* Was waiting, start a new cycle. */
            timeout_i2c = 0;

            i2c_device_scan = i2c_device_enable;

            do_i2c_load_scan_cmd(&i2c);

            /* New cycle -> write at the start of a new buffer */
            i2c.boff = 0;
            i2c.state = I2C_CMD_NEXT;

            /* Scanning */
            i2c_bus_status = I2C_SCAN_RUN;

        } else if (i2c_bus_status == I2C_SCAN_RESET && timeout_i2c++ == 100000) {

            /* Was in reset (power-off), re-enable monimod */
            if (REGS->monimod_shdn != 0) {
                REGS->monimod_shdn = 0;
                /* Need to run the init sequence again. */
                i2c_device_reset |= I2C_SLV_ADC;
            }

            if (REGS->fan_ctrl != 0) {
                REGS->fan_ctrl = 0;
            }

            REGS->i2c_off = 0;
            i2c_bus_status = I2C_SCAN_ENABLE;
            timeout_i2c = 0;

        }

    }
}
