#!/usr/bin/env python3

import sys
import time
import subprocess
import socket
import random

class flags():
    verbose = False
    flag_dry = False
    flag_flash = False
    flag_iram_err = False
    flag_all = False
    flag_monitor = False
    flag_load = True
    log = None

shmem_addr = 0x401000

class CommError(Exception):
    pass

class FipUrvDrv:
    def __init__(self, trace=False):
        if False:
            self.proc = subprocess.Popen(["./mfip_urv", "-p"],
                                         universal_newlines=True,
                                         bufsize=0,
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE)
            self.sock = None
        else:
            self.proc = None
            self.sock = socket.create_connection(("127.0.0.1", 2000))

        self.trace = trace
        self.nbr_comm_err = 0
        self.nbr_ping_err = 0
        self.nbr_flash_err = 0

    def read_reply(self):
        l = ''
        res = []
        while True:
            if self.sock:
                c = self.sock.recv(1).decode('ascii')
            else:
                c = self.proc.stdout.read(1)
            if c == 0:
                # EOF
                break
            if c == '\n':
                if len(l) > 2 and l[:2] == '# ':
                    break
                res.append(l)
                l = ''
            else:
                l += c
        if self.trace:
            sys.stderr.write("Recv: {} + {}\n".format(res, l))
        return (res, l)

    def send_cmd(self, cmd):
        if self.trace:
            sys.stderr.write("Send: {}\n".format(cmd))
        if self.sock:
            self.sock.send(cmd.encode('ascii') + b'\n')
        else:
            self.proc.stdin.write(cmd + '\n')


    def cmd(self, cmd):
        """Send a command and return the reply"""
        self.send_cmd(cmd)
        res, last = self.read_reply()
        if last != "# OK":
            print("ERROR: comm, status: {}".format(last))
            self.nbr_comm_err += 1
            raise CommError
        return res

    def cmd_crash(self, cmd):
        """Send a command but expect a timeout"""
        self.send_cmd(cmd)
        res, last = self.read_reply()
        if last == "# OK":
            print("ERROR: comm failure expected")
            raise CommError

    def disp(self, res):
        for l in res:
            print(l, end='')

    def close(self):
        if self.proc:
            # Close the sub-process
            self.proc.stdin.close()
            self.read_reply()
            self.proc.wait()
        else:
            self.sock.close()


def get_pow_vc(s, page):
    s.cmd("wmon page {}".format(page))
    vout = s.cmd("mon -s vout")
    iout = s.cmd("mon -s iout")
    return (vout[0], iout[0])

def get_pow_v(s, page):
    s.cmd("wmon page {}".format(page))
    vout = s.cmd("mon -s vout")
    return vout[0]


def get_cmd(s, cmd):
    res = s.cmd("mon -s {}".format(cmd))
    return res[0]

def show_cmd(s, cmd, name):
    res = get_cmd(s, cmd)
    print("{}:  {}".format(name, res))

REGS = 0x100000
SPI = REGS + 0xe0

def spi_init(s):
    # Set divider (1Mhz)
    s.cmd("write 0x{:x} 20".format(SPI + 0x14))

def spi_cmd(s, cmd):
    """Send cmd to spi.  cmd is a list of bytes"""
    l = len(cmd)
    assert 0 < l <= 8

    # Set ctrl char-len=5*8=0x28, tx_neg=1, rx_neg=0
    s.cmd("write 0x{:x} 0x{:04x}".format(SPI + 0x10, 0x400 + l * 8))

    # Bytes to be sent
    ra = SPI + ((l - 1) // 4) * 4

    w = 0
    r = ra
    for i in range(l):
        off = (l - 1 - i) & 3
        #print("  l={}, i={}, off={}".format(l, i, off))
        w |= cmd[i] << (off * 8)
        if off == 0 or i == l - 1:
            s.cmd("write 0x{:x} 0x{:x}".format(r, w))
            r -= 4
            w = 0

    # Set SS=1
    s.cmd("write 0x{:x} 1".format(SPI + 0x18))

    # Go
    s.cmd("write 0x{:x} 0x{:04x}".format(SPI + 0x10, 0x500 + l * 8))

    # Read
    res = []
    r = ra
    for i in range(l):
        off = (l - 1 - i) & 3
        if i == 0 or off == 3:
            w = s.cmd("read 0x{:x}".format(r))
            r -= 4
            w = int(w[0], 0)
        v = (w >> (off * 8)) & 0xff
        res.append(v)

    # Set SS=0
    s.cmd("write 0x{:x} 0".format(SPI + 0x18))

    return res

def flash_id(s):
    spi_init(s)

    res = spi_cmd(s, [0x05, 0])
    print("Flash status: {:02x}".format(res[1]))

    res = spi_cmd(s, [0x9f, 0, 0, 0, 0])
    print("Flash id: {:02x}{:02x}{:02x}{:02x}".format(
        res[1], res[2], res[3], res[4]))

    # Read
    for i in range(10):
        res = spi_cmd(s, [0x03, 0, i, 0, 0, 0, 0, 0])
        print("Flash 0x{:04x}: {:02x} {:02x} {:02x} {:02x}".format(
            i << 8, res[4], res[5], res[6], res[7]))

def flash_wait_write(s):
    while True:
        res = spi_cmd(s, [0x05, 0])
        print("Flash status: {:02x}".format(res[1]))
        if (res[1] & 0x20) != 0:
            print("ERROR: flash write/erase")
            s.nbr_flash_err += 1
            break
        if (res[1] & 1) == 0:
            break

def flash_prg(s, pg, off):
    print("Write flash at page={:02x} off={:02x}".format(pg, off))

    # Write enable
    spi_cmd(s, [0x06])

    # Write
    npg = (pg + 1) & 0xff
    spi_cmd(s, [0x02, 0x00, pg, off, pg, (~pg) & 0xff, npg, (~pg) & 0xff])

    flash_wait_write(s)

    # Write disable
    spi_cmd(s, [0x04])


def flash_full_erase(s):
    print("Erasing flash")

    # Write enable
    spi_cmd(s, [0x06])

    # Erase
    spi_cmd(s, [0x60])

    flash_wait_write(s)

def flash_page_erase(s, pg):
    print("Erasing flash page 0x{:02x}".format(pg))

    # Write enable
    spi_cmd(s, [0x06])

    # Page erase
    spi_cmd(s, [0x81, 0x00, pg, 0x00])

    flash_wait_write(s)

def flash_find_free(s):
    """Assuming flash pages are written consecutively, find the first erased
    one"""
    first = 0
    last = 255

    print("Find first free page:")
    while True:
        mid = (first + last) // 2
        res = spi_cmd(s, [0x03, 0, mid, 0, 0, 0])
        print("Page {}: first byte is {:02x} {:02x} (first:{}, last:{})".format(
            mid, res[4], res[5], first, last))
        if res[4] == 0xff and res[5] == 0xff:
            # Page is free
            if mid == first:
                return mid
            last = mid - 1
        else:
            # Page is busy
            if mid == last:
                if last == 255:
                    return -1
                return last + 1
            first = mid + 1

def show_psu(s):
    psu = s.cmd("psu")
    s.cmd("i2c 2")
    print("********** PSU 0 ******************")
    if psu[0] == "PSU 0: presence: 0":
        print("Not present")
    elif psu[0] == "PSU 0: presence: 1":
        print("TODO")
    else:
        print("unhandled")
    print("********** PSU 1 ******************")
    if psu[1] == "PSU 1: presence: 0":
        print("Not present")
    elif psu[1] == "PSU 1: presence: 1":
        show_cmd(s, "-t 0x18 vout", "Vout")
        show_cmd(s, "-t 0x18 iout", "Iout")
        show_cmd(s, "-t 0x18 temp-1", "Temp 1")
        show_cmd(s, "-t 0x18 temp-2", "Temp 2")
    else:
        print("unhandled")

def show_boards(s):
    print("********** Configuration *********")
    s.cmd("i2c 1")
    print("SLOTS: 01 02 03 04 05 06 07 08")
    print("      ", end='')
    for i in range(8):
        res = s.cmd("i2c-id -s {}".format(i))
        print(" {}".format(res[0]), end='')
    print()

def show_sysmon(s):
    s.cmd("i2c 0")
    t = time.time()
    ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
    rev = get_cmd(s, "revision")
    print("******** System board (revision {}) ********".format(rev))
    print("time: {}".format(ts))
    p12 = get_pow_vc(s, 0)
    print("P12V:   {}   {}".format(p12[0], p12[1]))
    p1_2 = get_pow_v(s, 1)
    print("P1.2V:  {}".format(p1_2))
    p3_3 = get_pow_v(s, 2)
    print("P3.3V:  {}".format(p3_3))
    pperi = get_pow_v(s, 3)
    print("Vperi:  {}".format(pperi))
    tamb = get_cmd(s, "temp-1")
    print("Temp amb:  {}".format(tamb))
    tldo = get_cmd(s, "temp-2")
    print("Temp LDO:  {}".format(tldo))
    tfpga = get_cmd(s, "temp-3")
    print("Temp FPGA: {}".format(tfpga))
    print("CSV: {tm},{p12v},{p12i},{p1_2},{p3_3},{tamb},{tldo},{tfpga},{ncom_err},{nflash_err},{nping_err}".format(
        tm=ts,
        p12v=p12[0].rstrip(' V'),
        p12i=p12[1].rstrip(' A'),
        p1_2=p1_2.rstrip(' V'),
        p3_3=p3_3.rstrip(' V'),
        tamb=tamb.rstrip(' C'),
        tldo=tldo.rstrip(' C'),
        tfpga=tfpga.rstrip(' C'),
        ncom_err=s.nbr_comm_err,
        nflash_err=s.nbr_flash_err,
        nping_err=s.nbr_ping_err))

rnd = random.Random()

def do_ping(s):
    v = rnd.randint(0, (1<<32) - 1)
    print("Ping with value {:08x}".format(v))
    s.cmd("write 0x107fc 0x{:08x}".format(v))
    w = s.cmd("read 0x107fc")
    rv = int(w[0], 0)
    if rv != v:
        print("ERROR: ping (write {:08x}, read {:08x})".format(v, rv))
        s.nbr_ping_err += 1

def extract(s, left, right=None):
    if len(s) < len(left):
        sys.stderr.write("extract ERR1: '{}' from '{}'\n".format(left, s))
        return "ERR1"
    if s[:len(left)] != left:
        sys.stderr.write("extract ERR2: '{}' from '{}'\n".format(left, s))
        return "ERR2"
    s = s[len(left):]
    if right is None:
        return s
    if len(s) < len(right):
        sys.stderr.write("extract ERR3: '{}' from '{}'\n".format(left, s))
        return "ERR3"
    if s[-len(right):] != right:
        sys.stderr.write("extract ERR4: '{}' from '{}'\n".format(left, s))
        return "ERR4"
    return s[:-len(right)]

class tiny_tunit():
    """A very-very light pytest.  Simpler to use, less dependencies"""
    def __init__(self):
        self.nbr_tests = 0
        self.nbr_errors = 0
        self.total_errors = 0
        self.started = False

    def start(self, msg):
        self.failif(self.started, "previous test not done")
        self.nbr_tests += 1
        print(msg, '... ', end='')
        self.started = True

    def fail(self, msg):
        self.nbr_errors += 1
        print('ERROR: {}'.format(msg))

    def failif(self, cond, msg):
        if cond:
            self.fail(msg)

    def failifneq_dec(self, val, expected, msg):
        if val != expected:
            self.fail("{} (got {}, expect {})".format(msg, val, expected))

    def done(self):
        self.failif(not self.started, "test not started")
        if self.nbr_errors == 0:
            print ('OK')
        else:
            print('KO')
        self.total_errors += self.nbr_errors
        self.nbr_errors = 0
        self.started = False

    def report(self):
        print('Nbr tests: {}'.format(self.nbr_tests))
        if self.total_errors == 0:
            print('PASS')
            return True
        else:
            print('FAILED: {} error(s)'.format(self.total_errors))
            return False

class mgmt_tests(tiny_tunit):
    def __init__(self, s):
        super().__init__()
        self.s = s

    def cmd(self, c):
        """Send command :param c:"""
        return self.s.cmd(c)

    def do_extract_recovery(self):
        p = self.cmd("stat-regs")
        nrecv = int(extract(p[4], "nbr recovery: "), 0)
        nrst = int (extract(p[5], "nbr resets: "), 0)
        last = int(extract(p[6], "last reset: "), 0)
        return (nrecv, nrst, last)

    def do_extract_reset_cause(self):
        p = self.cmd("supervisor")
        return int(extract(p[0], "reset cause: "), 0)

    def t_recovery(self):
        self.start("check mgmt CPU recovery")
        nrecv, nrst, _ = self.do_extract_recovery()
        for i in range(3):
            rst_cause = self.do_extract_reset_cause()
            #print("Fault on cpu {}".format(i))
            self.s.cmd_crash("cpu-fault {}".format(i))
            nrecv1, nrst1, last = self.do_extract_recovery()
            self.failif(nrecv1 != nrecv + 1,
                        "Bad nbr recovery: read {}, expect {}".format(
                            nrecv1, nrecv + 1))
            nrecv = nrecv1
            self.failif(nrst1 != nrst + 1,
                        "Bad nbr reset: read {}, expect {}".format(
                            nrst1, nrst + 1))
            nrst = nrst1
            self.failif(last != (1 << i),
                        "Bad last reset: read {}, expect {}".format(last, 1 << i))
            if last != (1 << i):
                cause = self.do_extract_reset_cause()
                self.failif(cause != 1,
                            "bad reset cause, read {}, expect 1".format(cause))
        self.done()

    def do_extract_dram_ecc_from_out(self, p):
        cyc = int(extract(p[5], "IRAM scrub cycle number: "), 0)
        n_ecc_se = int(extract(p[11],
                               "DRAM corrected errors by scrubber: "), 0)
        n_ecc_de = int(extract(p[13],
                               "DRAM double errors detected in last cycle: "), 0)
        return cyc, n_ecc_se, n_ecc_de

    def do_extract_dram_ecc(self):
        p = self.cmd("supervisor")
        return self.do_extract_dram_ecc_from_out(p)

    def t_dram_se(self):
        self.start("check mgmt DRAM single ecc")
        p = self.cmd("config")
        sz = int(extract(p[3], "mgmt dram size: "), 0)
        self.failif(sz < 1024,
                    "bad dram size ({})".format(sz))
        if sz < 1024:
            return
        cyc, n_ecc_se, n_ecc_de = self.do_extract_dram_ecc()
        # print("se: {}, de: {} (before the test)".format(n_ecc_se, n_ecc_de))
        self.failif(n_ecc_de != 0,
                    "Already data ECC errors")
        # Poison for a single error
        self.cmd("write-ecc 0x403f0 0x123 1")
        # Check single error correction
        cyc1, n_ecc_se1, n_ecc_de1 = self.do_extract_dram_ecc()
        #print("se: {}, de: {} (after single error)".format(n_ecc_se1, n_ecc_de1))
        self.failif(n_ecc_se1 != n_ecc_se + 1 or n_ecc_de1 != 0,
                    "bad number of ECC errors")
        self.done()

    def t_dram_de_scrub_off(self, off=0x3f4):
        self.start("check mgmt DRAM double ecc by scrubber (at 0x{:x})".format(
            off))
        addr = 0x40000 + off
        _, n_ecc_se, n_ecc_de = self.do_extract_dram_ecc()
        # Poison for a double error
        self.cmd("write-ecc 0x{:x} 0x456 6".format(addr))
        cyc, _, _ = self.do_extract_dram_ecc()
        p1 = self.cmd('supervisor')
        cyc1, n_ecc_se1, n_ecc_de1 = self.do_extract_dram_ecc_from_out(p1)
        self.failif(cyc == cyc1, "no full scrub cycle (0x{:x} and 0x{:x}".format(cyc, cyc1))
        #print("se: {}, de: {} (after double error)".format(n_ecc_se1, n_ecc_de1))
        self.failifneq_dec(n_ecc_se1, n_ecc_se, "bad number of se ECC errors")
        self.failifneq_dec(n_ecc_de1, 1, "bad number of de ECC errors")
        self.cmd("write-ecc 0x{:x} 0 0".format(addr))
        self.done()

    def t_dram_de_scrub_scan(self):
        for off in (0x3f0, 0x3f4, 0x3f8, 0x3fc, 0x0, 0x04):
            self.t_dram_de_scrub_off(off)

    def t_dram_de(self):
        self.start("check mgmt DRAM double ecc")
        addr = 0x40000 + 0x3f4
        _, n_ecc_se, n_ecc_de = self.do_extract_dram_ecc()
        # Poison for a double error
        self.cmd("write-ecc 0x{:x} 0x456 6".format(addr))
        cyc, _, _ = self.do_extract_dram_ecc()
        p1 = self.cmd('supervisor')
        cyc1, n_ecc_se1, n_ecc_de1 = self.do_extract_dram_ecc_from_out(p1)
        self.failif(cyc == cyc1, "no full scrub cycle (0x{:x} and 0x{:x}".format(cyc, cyc1))
        #print("se: {}, de: {} (after double error)".format(n_ecc_se1, n_ecc_de1))
        self.failifneq_dec(n_ecc_se1, n_ecc_se, "bad number of se ECC errors")
        self.failifneq_dec(n_ecc_de1, 1, "bad number of de ECC errors")
        if n_ecc_de1 == 0:
            print ("cyc0: 0x{:x}, cyc1: 0x{:x}".format(cyc, cyc1))
            print(p1)
            print(self.cmd('supervisor'))
        # Read the double error (to trigger a reset)
        self.s.cmd_crash("read 0x{:x}".format(addr))
        cause = self.do_extract_reset_cause()
        self.failif(cause != 4,
                    "bad reset cause, read {}, expect 4".format(cause))
        self.cmd("write-ecc 0x{:x} 0 0".format(addr))
        self.done()

    def do_extract_iram_ecc(self):
        p = self.cmd("supervisor")
        n_ecc_se = int(extract(p[6], "IRAM corrected errors by scrubber: "), 0)
        n_ecc_de = int(extract(p[8],
                               "IRAM double errors detected in last cycle: "), 0)
        return (n_ecc_se, n_ecc_de)

    def t_iram(self):
        self.start("check mgmt IRAM ecc")
        p = self.cmd("config")
        sz = int(extract(p[2], "mgmt iram size: "), 0)
        self.failif(sz < 1024,
                    "bad iram size ({})".format(sz))
        if sz < 1024:
            return
        n_ecc_se, n_ecc_de = self.do_extract_iram_ecc()
        #print("se: {}, de: {} (before the test)".format(n_ecc_se, n_ecc_de))
        self.failif(n_ecc_de != 0,
                    "Already insn ECC errors")
        # Request to poison IRAM on the next restart
        self.cmd("restart-cfg 7")
        # So need to restart (by doing a double dram error)
        #print("crash...")
        self.cmd("write-ecc 0x403f4 0x456 6")
        self.s.cmd_crash("read 0x403f4")
        # Will timeout
        cause = self.do_extract_reset_cause()
        self.failif(cause != 4,
                    "bad reset cause, read {}, expect 4".format(cause))
        # Restore restart config (to not poison IRAM anymore)
        self.cmd("restart-cfg 3")
        n_ecc_se1, n_ecc_de1 = self.do_extract_iram_ecc()
        #print("se: {}, de: {} (after test)".format(n_ecc_se1, n_ecc_de1))
        # The IRAM poison is 2 se and 2 de.
        self.failif(n_ecc_se1 != n_ecc_se + 2 or n_ecc_de1 != 2,
                    "Error in IRAM ecc")
        # Crash again with a double error
        self.s.cmd_crash("read 0xfffc")
        cause = self.do_extract_reset_cause()
        self.failif(cause != 2,
                    "bad reset cause, read {}, expect 2".format(cause))
        self.done()

    def t_watchdog(self):
        self.start("check mgmt watchdog")
        nrecv, nrst, _ = self.do_extract_recovery()
        #print("Trigger watchdog")
        self.s.cmd_crash("wait 2000 0")
        _, nrst1, _ = self.do_extract_recovery()
        self.failif(nrst1 != nrst + 1,
                    "Bad nbr reset: read {}, expect {}".format(
                        nrst1, nrst + 1))
        cause = self.do_extract_reset_cause()
        self.failif(cause != 8,
                    "bad reset cause, read {}, expect 8".format(cause))
        self.done()

    def run_all_tests(self):
        self.t_recovery()
        self.t_watchdog()
        self.t_dram_se()
        self.t_dram_de_scrub_scan()
        self.t_dram_de()
        self.t_iram()
        return self.report()


def do_extract_monimod_cycle(s):
    p = s.cmd("monimod-read")
    return int(extract(p[0], "Cycle: "), 0)

def do_extract_monimod_reset(s):
    p = s.cmd("stat-regs")
    return int(extract(p[0], "nbr monimod reset: "), 0)

def do_extract_mbox1(s):
    p = s.cmd("mbox 1")
    if flags.verbose:
        print("mbox 1: {}".format(p))
    sta = int(extract(p[0], "mailbox status    (@0x004001b8): "), 0)
    val = int(extract(p[1], "mailbox in value  (@0x004001b4): "), 0)
    return sta, val

def do_self_tests_monimod_scan(s, cnt=20):
    cyc = do_extract_monimod_cycle(s)
    ntry = cnt
    while ntry > 0:
        ncyc = do_extract_monimod_cycle(s)
        if ncyc != cyc:
            break
        ntry -= 1
    if ntry == 0:
        print("no monimod scan")

def do_self_tests_monimod(s):
    print('* Test monimod')
    do_self_tests_monimod_scan(s)
    nrst = do_extract_monimod_reset(s)
    s.cmd("monimod-power off")
    do_self_tests_monimod_scan(s, 100)
    nrst1 = do_extract_monimod_reset(s)
    if nrst1 != nrst + 1:
        print("monimod reset expected")

def do_mgmt_tests(s):
    mgmt = mgmt_tests(s)
    mgmt.run_all_tests()
    print("self tests done")


class app_common_tests(tiny_tunit):
    def __init__(self, s, idx):
        super().__init__()
        self.s = s
        self.idx = idx

    def cmd(self, c):
        """Send command :param c:"""
        return self.s.cmd(c)

    def common_reset(self):
        self.s.cmd('plc {} reset'.format(self.idx))
        self.s.cmd('plc {} iwrite'.format(self.idx))

    def reset(self):
        # Default implementation
        self.common_reset()

    def load(self, filename):
        print('loading app{} test app'.format(self.idx))
        self.common_reset()
        self.s.cmd('load {} {}'.format(self.idx, filename))

    def run(self):
        self.s.cmd('plc {} run'.format(self.idx))

    def restart(self):
        self.reset()
        self.run()

    def get_state(self):
        p = self.s.cmd('plc {}'.format(self.idx))
        rst = int(extract(p[0], "plc reset:  ", "  # 1: sys-rst, 2: cpu-rst"))
        status = int(extract(p[1], "plc status: ", "  # 1: fault"))
        return rst, status

    def get_supervisor_rst(self):
        p = self.s.cmd('plc-supervisor -s {} reset-cause cpu-reset'.format(self.idx))
        cause = int(extract(p[0], "reset-cause: "), 0)
        nbr_rst = int(extract(p[1], "cpu-reset: "), 0)
        return cause, nbr_rst

    def get_supervisor_flt(self):
        p = self.s.cmd('plc-supervisor -s {} reset-cause user-fault'.format(self.idx))
        cause = int(extract(p[0], "reset-cause: "), 0)
        nbr_flt = int(extract(p[1], "user-fault: "), 0)
        return cause, nbr_flt

    def get_dram_errors(self):
        p = self.s.cmd('plc-supervisor {}'.format(self.idx))
        corr = int(extract(p[5], "PLC DRAM corrected errors: "))
        scrub = int(extract(p[7], "PLC DRAM corrected errors by scrubber: "))
        dble = int(extract(p[8], "PLC DRAM double errors detected in last cycle: "))
        return corr, scrub, dble

    def get_iram_errors(self):
        p = self.s.cmd('plc-supervisor {}'.format(self.idx))
        corr = int(extract(p[1], "PLC IRAM corrected errors: "))
        scrub = int(extract(p[3], "PLC IRAM corrected errors by scrubber: "))
        dble = int(extract(p[4], "PLC IRAM double errors detected in last cycle: "))
        return corr, scrub, dble

    def poison_tfunc(self, ret, ecc):
        assert(0 <= ret <= 0xfff)
        self.cmd('write-plc-ecc {} 0x10 0x{:03x}00513 0x{:x}'.format(
            self.idx, ret, ecc))

    def poison_iram(self, ret, ecc):
        self.reset()
        self.poison_tfunc(ret, ecc)
        self.run()

    def enable_iram_scrubber(self):
        self.enable_scrubber(1)

    def enable_dram_scrubber(self):
        self.enable_scrubber(2)

    def t_iram_se_cpu(self):
        self.start('check iram single-error (by cpu)')
        self.poison_iram(0x3d, 1)
        corr, scrub, dble = self.get_iram_errors()
        self.failif(corr != 0 or scrub != 0 or dble != 0,
                    'IRAM ECC errors already present')
        self.simple_data_transfer(0x11112222)
        v = self.call_tfunc()
        self.failif(v != 0x3d,
                    'incorrect data (got {:08x} expect {:08x})'.format(v, 2))
        corr, scrub, dble = self.get_iram_errors()
        self.failif(corr != 1 or scrub != 0 or dble != 0,
               'IRAM ECC errors must have been detected')
        self.done()

    def t_iram_de_cpu(self):
        self.start('check iram double-error (by cpu)')
        self.poison_iram(0x10, 0x11)
        corr, scrub, dble = self.get_iram_errors()
        self.failif(corr != 0 or scrub != 0 or dble != 0,
                    'IRAM ECC errors already present')
        self.simple_data_transfer(0x11112222)
        self.call_tfunc_crash()
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x02,
              'plc fault (iram ecc) expected (got cause: 0x{:x})'.format(cause))
        # Un-poison iram
        self.poison_iram(0x22, 0)
        v = self.call_tfunc_recov()
        self.failif(v != 0x22,
                    "incorrect value returned ({:x})".format(v))
        self.done()

    def t_iram_se_scrub(self):
        self.start('check iram single-error (by scrubber)')
        # Poison iram
        self.poison_iram(0xa5, 4)
        # Enable iram scrubber
        self.enable_iram_scrubber()
        self.simple_data_transfer(0x11224433)
        corr, scrub, dble = self.get_iram_errors()
        self.failif(corr != 0 or scrub != 1 or dble != 0,
                    'IRAM ECC errors must have been detected')
        self.done()

    def t_iram_de_scrub(self):
        self.start('check iram double-error (by scrubber)')
        # Poison iram
        self.poison_iram(0xc1, 0x48)
        # Enable iram scrubber
        self.enable_iram_scrubber()
        self.simple_data_transfer(0x33334444)
        corr, scrub, dble = self.get_iram_errors()
        self.failif(corr != 0 or scrub != 0 or dble != 1,
                    'IRAM ECC errors must have been detected')
        # Un-poison iram
        self.poison_iram(0x01, 0x00)
        self.done()

    def t_iram_de_scrub_scan(self):
        for off in (0xff0, 0xff4, 0xff8, 0xffc):
            self.start('check iram double-error at 0x{:x} (by scrubber)'.format(
                off))
            # Poison iram
            self.reset()
            self.cmd('write-plc-ecc {} 0x{:x} 0x00 0x{:x}'.format(
                self.idx, off, 0x12))
            self.run()
            # Enable iram scrubber
            self.enable_iram_scrubber()
            self.simple_data_transfer(0x33334444)
            corr, scrub, dble = self.get_iram_errors()
            self.failif(corr != 0 or scrub != 0 or dble != 1,
                        'IRAM ECC errors must have been detected')
            # Un-poison iram
            self.reset()
            self.cmd('write-plc-ecc {} 0x{:x} 0x00 0x{:x}'.format(
                self.idx, off, 0))
            self.run()
            self.done()

    def t_dram_se_cpu(self):
        self.start('check dram single error (by cpu)')
        corr, scrub, dble = self.get_dram_errors()
        self.failif(corr != 0 or scrub != 0 or dble != 0,
                    'ECC errors already present')
        # Write a poisonned value
        v0 = self.poison_mask(0x84b35e0f)
        self.poison_scratch(0x1, v0)
        # ... and read it
        v = self.read_scratch()
        self.failif(v != v0,
               'value read is incorrect (got {:08x}, expect {:08x})'.format(v, v0))
        corr, scrub, dble = self.get_dram_errors()
        self.failif(corr != 1 or scrub != 0 or dble != 0,
                    'Error must have been detected')
        # Read again...
        v = self.read_scratch()
        self.failif(v != v0,
                    'value read is incorrect (got {:08x}, expect {:08x})'.format(v, v0))
        # but do not expect another error!
        corr, scrub, dble = self.get_dram_errors()
        self.failif(corr != 1 or scrub != 0 or dble != 0,
                    'Error must have been corrected')
        self.done()

    def t_dram_de_cpu(self):
        self.start('check dram double error (by cpu)')
        v0 = self.poison_mask(0xdcba5432)
        self.poison_scratch(0x5, v0)
        # Read the memory, but do not expect any reply (as it will go faulty)
        self.read_scratch_crash()
        rst, status = self.get_state()
        self.failif(status != 1, 'plc should be in fault state')
        # cpu is now stuck in faulty state, need to reset and restart it
        self.restart()
        # purge the aborted read_scratch
        self.read_scratch_recover()
        v = self.read_scratch()
        self.failif(v != 0,
               'value read is incorrect (got {:08x}, expect 0)'.format(v))
        self.done()

    def t_dram_se_scrub(self):
        self.start('check dram single-error (by scrubber)')
        # Enable data scrubber
        self.enable_dram_scrubber()
        v0 = self.poison_mask(0x4680feab)
        self.poison_scratch(0x8, v0)
        corr, scrub, dble = self.get_dram_errors()
        self.failif(corr != 0 or scrub != 1 or dble != 0,
                    'Scrubber must have detected errors')
        self.done()

    def t_dram_de_scrub(self):
        self.start('check dram double-error (by scrubber)')
        self.poison_scratch(0x6, 0)
        corr, scrub, dble = self.get_dram_errors()
        self.failif(corr != 0 or dble != 1,
                    'Scrubber must have detected errors')
        self.poison_scratch(0, 0) # un-poison memory
        self.done()

    def t_stop_restart(self, num):
        self.start('check stop and restart #{}'.format(num))
        self.reset()
        self.run()
#        print(plc.s.cmd('mbox 0'))
        v = num + (rnd.randint(0, 0xffff) << 16)
        self.simple_data_transfer(v)
        self.done()

    def t_watchdog(self):
        self.start('check watchdog')
        self.disable_watchdog_keepalive()
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x08,
                    'plc bad reset cause (got 0x{:x}, expect 0x08)'.format(cause))
        self.failif(nbr_rst != 1, 'unexpected reset for plc')
        rst, status = self.get_state()
        self.failif(status != 1, 'plc should be in fault state')
        self.restart()
        self.done()


class app0_tests(app_common_tests):
    def __init__(self, s):
        super().__init__(s, 0)
        self.counter = 1 # packet counter from app
        self.mgmt_count = None # packet counter from mgmt
        self.nbr_recov = 0

    def load(self):
        super().load('plc0_autotest.elf')
        self.nbr_recov = 0

    def reset(self):
        self.common_reset()
        self.counter = 1
        self.nbr_recov = 0

    def do_last_data(self):
        p = self.s.cmd('last-data')
        if flags.verbose:
            print('last-data: {}'.format(p))
        cnt = int(extract(p[0], 'count: '))
        data = extract(p[1], 'data: ').split()
        if self.mgmt_count is not None:
            if cnt == self.mgmt_count:
                self.fail('no data packet received')
                return data
            elif cnt != self.mgmt_count + 1:
                self.fail('bad mgmt packet counter - data packet missed')
        self.mgmt_count = cnt
        counter = int(data[4], 16)
        self.failif(self.counter is not None and counter != self.counter,
                    'bad count number (got {}, expect {})'.format(counter, self.counter))
        self.counter = (counter + 1) & 0xff
        return data

    def send_data_only(self, dat):
        if flags.verbose:
            print('data: {}'.format(dat))
        self.s.cmd('data ' + dat)

    def send_data(self, dat):
        self.send_data_only(dat)
        # Wait a little bit to receive the data
        self.s.cmd('ping')
        return self.do_last_data()

    def num2bytes(self, v):
        b = v.to_bytes(4, 'little')
        return ' '.join(['0x{:02x}'.format(c) for c in b])

    def bytes2num(self, v):
        b = [int(v[i], 16) for i in range(0,4)]
        return int.from_bytes(b, byteorder='little')

    def simple_data_transfer(self, v):
        data = self.send_data('0 0 0 0 {}'.format(self.num2bytes(v)))
        edata = self.bytes2num(data[8:12])
        nv = ~v & 0xffffffff
        self.failif(edata != nv,
                    'incorrect data (got {:08x} expect {:08x})'.format(edata, nv))

    def poison_scratch(self, ecc, v):
        self.send_data('0x{:02x} 0x0d 0x1c 0xa1 {}'.format(ecc, self.num2bytes(v)))

    def poison_mask(self, v):
        return v

    def read_my_reg(self, off):
        data = self.send_data('00 0x05 0x1c 0xa1 {} 0 0 0'.format(off))
        return self.bytes2num(data[8:12])

    def write_my_reg(self, off, val):
        self.send_data('0x{:02x} 0x08 0x1c 0xa1 {} 0 0 0'.format(off, self.num2bytes(val)))

    def read_supervisor_cpu(self):
        data = self.send_data('00 0x09 0x1c 0xa1 0 0 0 0')
        return self.bytes2num(data[8:12])

    def read_recovery(self):
        return self.read_my_reg(1)

    def read_last_reset(self):
        return self.read_my_reg(2)

    def read_scratch(self):
        return self.read_my_reg(4)

    def call_tfunc(self):
        p = self.send_data('00 07 0x1c 0xa1 0 0 0 0')
        return self.bytes2num(p[8:12])

    def call_tfunc_crash(self):
        """Call tfunc() but expect a crash (poisonned instruction)"""
        self.send_data_only('00 07 0x1c 0xa1 0 0 0 0')

    def call_tfunc_recov(self):
        # plc0 will re-execute the last command (the read) and reply.
        p = self.do_last_data()
        return self.bytes2num(p[8:12])

    def read_scratch_crash(self):
        self.send_data_only('00 0x05 0x1c 0xa1 {} 0 0 0'.format(4))

    def read_scratch_recover(self):
        # plc0 will re-execute the last command (the read) and reply.
        # there is no fault as the d-ram has been cleared during the reset.
        p = self.do_last_data()
        self.failif(p[8:12] != ['00', '00', '00', '00'],
                    "incorrect value read from memory ({})".format(p[8:12]))

    def enable_scrubber(self, v):
        self.send_data('00 0x0f 0x1c 0xa1 {} 0 0 0'.format(v))

    def disable_watchdog_keepalive(self):
        self.send_data('00 0x0c 0x1c 0xa1 0 0 0 0')

    def do_extract_mbox0(self):
        p = self.cmd("mbox 0")
        if flags.verbose:
            print("mbox 0: {}".format(p))
        sta = int(extract(p[0], "mailbox status    (@0x00400158): "), 0)
        ival = int(extract(p[1], "mailbox in value  (@0x00400154): "), 0)
        oval = int(extract(p[2], "mailbox out value (@0x00400150): "), 0)
        return sta, ival, oval

    def t_load(self):
        # Reset mbox so that there are no packets left
        _, ival, oval = self.do_extract_mbox0()
        v = ((ival & 0xff) << 8) | ((ival & 0xff00) >> 8)
        if v != (oval & 0xffff):
            self.cmd('mbox 0 0x{:x}'.format(v))
        self.load()

    def t_check_start(self):
        self.start('check app0 current state')
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0,
              'plc0 already got a reset (cause: 0x{:x})'.format(cause))
        self.failif(nbr_rst != 0, 'plc0 already got a reset')
        self.done()

    def t_exchange(self, val=0xc3a03cf5):
        self.start('check operation (simple data transfer)')
        # Also check the pkt_counter (should be 1)
        # If not 1, there was a packet left by a previous run (but mbox has
        # been cleared)
        #    print(plc.s.cmd('mbox 0'))
        self.simple_data_transfer(val)
        self.done()

    def t_double_divergence(self):
        self.start('check for double divergence')
        # Disable recovery
        self.write_my_reg(12, 1)
        # Read cpu
        val = self.read_supervisor_cpu()
        self.failif(val != 0x0,
                    'incorrect failing cpu (get {}, expect {})'.format(val, 0x00))
        # Divergence of cpu-2
        self.send_data('00 00 0x1c 0xa1 2 0 0 0')
        # Check no new recovery
        nrecov = self.read_recovery()
        self.failif(nrecov != self.nbr_recov,
                    'incorrect recovery number (get {}, expect {})'.format(nrecov, self.nbr_recov))
        # Check for no fault
        rst, status = self.get_state()
        self.failif(status != 0, 'plc should not be in fault state')
        # Check failing cpu (read SUPERVISOR->cpu)
        val = self.read_supervisor_cpu()
        self.failif(val != 0x2,
               'incorrect failing cpu (get {}, expect {})'.format(val, 0x02))
        # Second divergence
        #    print(plc.s.cmd('mbox 0'))
        self.send_data('00 00 0x1c 0xa1 4 0 0 0')
        rst, status = self.get_state()
        self.failif(status != 1, 'plc should be in fault state')
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x01,
                    'plc0 fault is incorrect (cause: 0x{:x})'.format(cause))
        #   print(plc.s.cmd('mbox 0'))
        self.restart()
        self.done()

    def t_cpu_div_insn(self, cpu):
        self.start('check for insn divergence (cpu-{})'.format(cpu + 1))
        # Do divergence
        self.send_data('00 00 0x1c 0xa1 {} 0 0 0'.format(1 << cpu))
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x10,
                   'plc0 bad reset cause (got 0x{:x}, expect 0x10)'.format(cause))
        self.failif(nbr_rst != 0, 'unexpected reset for plc0')
        self.nbr_recov += 1
        nrecov = self.read_recovery()
        self.failif(nrecov != self.nbr_recov,
                   'incorrect recovery number (get {}, expect {})'.format(nrecov, self.nbr_recov))
        reset_cpu = self.read_last_reset()
        self.failif(reset_cpu != (1 << cpu),
                   'incorrect cpu reset (get {})'.format(reset_cpu))
        self.done()

    def t_all_cpu_div_data(self):
        self.start('check for data divergence (all cpus)')
        self.send_data('00 01 0x1c 0xa1 3 0 0 0')
        rst, status = self.get_state()
        self.failif(status != 1, 'plc should be in fault state')
        self.restart()
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0,
              'plc0 already got a reset (cause: 0x{:x})'.format(cause))
        self.done()

    def t_cpu_div_data(self, cpu):
        self.start('check for data divergence (cpu-{})'.format(cpu + 1))
        p = self.send_data('00 01 0x1c 0xa1 {} 0 21 12'.format(1 << cpu))
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x10,
                   'expected reset from plc0 (cause: 0x{:x})'.format(cause))
        self.restart()
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0,
                   'plc0 already got a reset (cause: 0x{:x})'.format(cause))
        self.done()

    def t_cpu_div_addr_w(self, cpu):
        self.start('check for d-addr write divergence (cpu-{})'.format(cpu + 1))
        p = self.send_data('00 02 0x1c 0xa1 {} 0 21 12'.format(1 << cpu))
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x10,
                   'expected reset from plc0 (cause: 0x{:x})'.format(cause))
        self.restart()
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0,
                   'plc0 already got a reset (cause: 0x{:x})'.format(cause))
        self.done()

    def t_cpu_div_addr_r(self, cpu):
        self.start('check for d-addr read divergence (cpu-{})'.format(cpu + 1))
        p = self.send_data('00 03 0x1c 0xa1 {} 0 21 12'.format(1 << cpu))
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x10,
                   'expected reset from plc0 (cause: 0x{:x})'.format(cause))
        self.restart()
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0,
                   'plc0 already got a reset (cause: 0x{:x})'.format(cause))
        self.done()

    def t_cpu_div_be(self, cpu):
        self.start('check for byte-en divergence (cpu-{})'.format(cpu + 1))
        p = self.send_data('00 04 0x1c 0xa1 {} 0 0 0'.format(1 << cpu))
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0x10,
                   'expected reset from plc0 (cause: 0x{:x})'.format(cause))
        self.restart()
        cause, nbr_rst = self.get_supervisor_rst()
        self.failif(cause != 0,
                   'plc0 already got a reset (cause: 0x{:x})'.format(cause))
        self.done()

    def t_call_tfunc(self, expected_val):
        self.start('check operation (after restart due to watchdog)')
        v = self.call_tfunc()
        self.failif(v != expected_val,
                    'incorrect data (got {:08x} expect {:08x})'.format(
                        v, expected_val))
        self.done()

    def t_shmem_ecc_de(self):
        self.start('check app0 handling of double ecc errors')
        # Get shmem slot
        _, _, oval = self.do_extract_mbox0()
        wslot = (oval >> 8) & 0xff
        nslot = (wslot + 1) & 0x3
        addr = shmem_addr + (wslot << 7)
        # Manually write a pseudo-message, corrupt a word
        self.cmd("write 0x{:x} 0x00ff55aa".format(addr))
        self.cmd("shmem-ecc 0x22")
        self.cmd("write 0x{:x} 0x33ccee77".format(addr + 4))
        self.cmd("shmem-ecc 0")
        # Send the message
        noval = (oval & 0xffff00ff) | (nslot << 8)
        self.cmd("mbox 0 0x{:x}".format(noval))
        # Wait for the reply
        self.s.cmd('ping')
        p = self.do_last_data()
        # Check reply
        self.failif(p[1] != 'fd',
                    'detection of ecc error expected (got: {})'.format(p[1]))
        self.done()

    def run_all_tests(self):
        self.t_check_start()
        self.t_exchange()
        self.t_double_divergence()

        for i in range(0,2):
            self.t_stop_restart(i)

        #        if self.total_errors != 0:
        #            return

        self.t_iram_se_cpu()
        self.t_iram_de_cpu()

        self.t_iram_se_scrub()
        self.t_iram_de_scrub()
        self.t_iram_de_scrub_scan()

        for i in range(0,3):
            self.t_cpu_div_insn(i)

        self.t_all_cpu_div_data()

        for i in range(0,3):
            self.t_cpu_div_data(i)

        for i in range(0,3):
            self.t_cpu_div_addr_w(i)

        for i in range(0,3):
            self.t_cpu_div_addr_r(i)

        for i in range(0,3):
            self.t_cpu_div_be(i)

        self.t_watchdog()

        self.t_exchange(0x3e053ff0)

        self.t_dram_se_cpu()
        self.t_dram_de_cpu()

        self.t_dram_se_scrub()
        self.t_dram_de_scrub()

        self.t_call_tfunc(0x01)
        self.t_shmem_ecc_de()
        return self.report()


class app1_tests(app_common_tests):
    def __init__(self, s):
        super().__init__(s, 1)
        self.counter = 1 # packet counter from app
        self.mgmt_count = None # packet counter from mgmt
        self.nbr_recov = 0

    def load(self):
        super().load('plc1_autotest.elf')

    def do_mbox(self):
        return do_extract_mbox1(self.s)

    def do_mbox_read(self):
        sta, res = self.do_mbox()
        self.failif(sta != 1, "incorrect mbox state (got {}, expect 1)".format(sta))
        return res

    def t_start(self):
        self.start("check initial mbox after load")
        val = self.do_mbox_read()
        self.failif(val != 0x12345678, "incorrect initial value (got {:x}, expect 0x12345678)".format(val))
        self.done()

    def simple_data_transfer(self, v):
        v &= 0x7fffffff
        self.cmd("mbox 1 0x{:08x}".format(v))
        val = self.do_mbox_read()
        e = (~v) & 0xffffffff
        self.failif(val != e, "incorrect data value (got 0x{:x}, expect 0x{:x})".format(val, e))

    def req(self, op):
        self.cmd("mbox 1 0x{:08x}".format(op))

    def req_rep(self, op):
        self.req(op)
        return self.do_mbox_read()

    def call_tfunc(self):
        return self.req_rep(0x87000000)

    def call_tfunc_crash(self):
        self.req(0x87000000)

    def call_tfunc_recov(self):
        # tfunc has been called and crashed.  But as the mailbox has been
        # marked as read, the command is not re-executed.
        # To do like app0, recall it.
        return self.req_rep(0x87000000)

    def enable_scrubber(self, v):
        return self.req_rep(0x8f000000 | v)

    def poison_mask(self, v):
        return v & 0xffff

    def poison_scratch(self, ecc, v):
        self.req(0x8d000000 | (ecc << 16) | (v & 0x00ffff))

    def read_my_reg_req(self, off):
        self.req(0x85000000 | (off & 0xff))

    def read_my_reg(self, off):
        self.read_my_reg_req(off)
        return self.do_mbox_read()

    def read_scratch(self):
        return self.read_my_reg(4)

    def read_scratch_crash(self):
        self.read_my_reg_req(4)

    def read_scratch_recover(self):
        # Command has been read, do not expect a reply.
        pass

    def disable_watchdog_keepalive(self):
        return self.req_rep(0x8c000000)

    def t_exchange(self, val=0xa5013cf8):
        self.start("check app1 data transfer")
        self.simple_data_transfer(val)
        self.done()

    def check_supervisor_flt(self, ecause, enbr_flt):
        cause, nbr_flt = self.get_supervisor_flt()
        self.failif(cause != ecause,
                    'incorrect cause (got: 0x{:x}, expect: 0x{:x})'.format(cause, ecause))
        self.failif(nbr_flt != enbr_flt,
                    'incorrect nbr fault (got: {}, expect: {})'.format(nbr_flt, enbr_flt))

    def ecc_flip(self, cmd, v0):
        v = self.req_rep(cmd | v0)
        self.failif(v != v0,
                    'bad value from plc1 (got: 0x{:x}, expect: 0x{:x})'.format(v, v0))

    def t_corrupt_reg_ra(self):
        self.start('check app1 register ra ecc error')
        _, nbr_flt = self.get_supervisor_flt()
        self.ecc_flip(0x81000000, 0x2321)  # no ecc flip
        self.check_supervisor_flt(0, nbr_flt)
        self.ecc_flip(0x81010000, 0x7854)  # bank 0
        self.check_supervisor_flt(0, nbr_flt + 1)
#        self.ecc_flip(0x81020000, 0x2e6b)  # bank 1 (not used)
#        self.check_supervisor_flt(0, nbr_flt + 2)
        self.simple_data_transfer(0x44332211)
        self.done()

    def t_corrupt_reg_r0(self):
        self.start('check app1 register r0 ecc error')
        _, nbr_flt = self.get_supervisor_flt()
        self.ecc_flip(0x82000000, 0xbcde)  # no ecc flip
        self.check_supervisor_flt(0, nbr_flt)
        self.ecc_flip(0x82010000, 0xcdfe)  # bank 0
        self.check_supervisor_flt(0, nbr_flt + 1)
        self.ecc_flip(0x82020000, 0xef56)  # bank 1
        self.check_supervisor_flt(0, nbr_flt + 2)
        self.simple_data_transfer(0xfedcbebe)
        self.done()

    def t_corrupt_reg_bank0(self):
        self.start('check app1 register bank0 ecc error')
        _, nbr_flt = self.get_supervisor_flt()
        self.ecc_flip(0x83000000, 0xbcde)  # no ecc flip
        self.check_supervisor_flt(0, nbr_flt)
        self.ecc_flip(0x83010000, 0xcdfe)  # bank 0
        self.check_supervisor_flt(0, nbr_flt + 1)
        self.ecc_flip(0x83020000, 0xef56)  # bank 1 (not used)
        self.check_supervisor_flt(0, nbr_flt + 1)
        self.simple_data_transfer(0xfedcbebe)
        self.done()

    def t_corrupt_reg_bank1(self):
        self.start('check register bank1 ecc error')
        _, nbr_flt = self.get_supervisor_flt()
        self.ecc_flip(0x84000000, 0xbcde)  # no ecc flip
        self.check_supervisor_flt(0, nbr_flt)
        self.ecc_flip(0x84010000, 0xcdfe)  # bank 0 (not used)
        self.check_supervisor_flt(0, nbr_flt)
        self.ecc_flip(0x84020000, 0xef56)  # bank 1
        self.check_supervisor_flt(0, nbr_flt + 1)
        self.simple_data_transfer(0xfedcbebe)
        self.done()

    def t_corrupt_reg_fatal(self):
        self.start("check register double ecc error")
        _, nbr_flt = self.get_supervisor_flt()
        self.req(0x84030000 | 0xef56)  # bank 0 & 1
        rst, status = self.get_state()
        self.failif(status != 1, 'plc should be in fault state')
        self.check_supervisor_flt(1, nbr_flt)
        self.restart()
        self.done()


    def run_all_tests(self):
        self.t_exchange()

        for i in range(0,2):
            self.t_stop_restart(i)

        self.t_watchdog()

        self.t_corrupt_reg_ra()
        self.t_corrupt_reg_r0()
        self.t_corrupt_reg_bank0()
        self.t_corrupt_reg_bank1()
        self.t_corrupt_reg_fatal()

        self.t_iram_se_cpu()
        self.t_iram_de_cpu()
        self.t_iram_se_scrub()
        self.t_iram_de_scrub()
        self.t_iram_de_scrub_scan()

        self.t_dram_se_cpu()
        self.t_dram_de_cpu()
        self.t_dram_se_scrub()
        self.t_dram_de_scrub()
        return self.report()


def do_app0_tests(s, forever=False):
    plc = app0_tests(s)
    plc.t_load()

    while True:
        plc.run_all_tests()
        if plc.total_errors != 0:
            break
        if not forever:
            break


def do_app1_tests(s, forever=False):
    app = app1_tests(s)
    app.load()

    app.run_all_tests()
#    app.t_start()


class struct():
    pass

def do_shmem_supervisor(s):
    p = s.cmd("shmem-supervisor")
    res = struct()
    res.err = int(extract(p[0], "shmem corrected errors: "))
    res.mde = int(extract(p[1], "shmem uncorrected errors for mgmt: "))
    res.ade = int(extract(p[2], "shmem uncorrected errors for app: "))
    res.cyc = int(extract(p[3], "shmem scrub cycle number: "), 0)
    res.sce = int(extract(p[4], "shmem corrected errors by scrubber: "))
    res.sde = int(extract(p[5], "shmem double errors detected in last cycle: "))
    res.cfg = int(extract(p[6], "shmem scrub config: "))
    res.per = int(extract(p[7], "shmem scrub period: "))
    return res

class shmem_tests(tiny_tunit):
    def __init__(self, s):
        super().__init__()
        self.s = s

    def cmd(self, c):
        """Send command :param c:"""
        return self.s.cmd(c)

    def get_supervisor(self):
        return do_shmem_supervisor(self.s)

    def t_init(self):
        self.start("check shmem initial state")
        sv = self.get_supervisor()
        self.failif(res.err != 0,
                    "shmem must have no errors")
        self.failif(res.sce != 0,
                    "shmem scrubber corrected errors must be 0")
        self.failif(res.sde != 0,
                    "shmem double errors must be 0")
        self.failif(res.cfg != 0,
                    "shmem scrubber must not be enabled")
        self.failif(res.per != 2047,
                    "shmem scrubber period is not correct")
        self.done()

    def do_read(self, off):
        p = self.cmd('read 0x{:x}'.format(shmem_addr + off))
        return int(p[0], 0)

    def do_write(self, off, val):
        self.cmd('write 0x{:x} 0x{:x}'.format(shmem_addr + off, val))

    def do_rd_cmp(self, off, v0):
        v = self.do_read(off)
        self.failif(v != v0,
                    "shmem incorrect read value (got 0x{:x}, expect 0x{:x})".format(v, v0))

    def do_rw(self, off, v0):
        self.do_write(off, v0)
        self.do_rd_cmp(off, v0)

    def t_rw(self):
        self.start("check shmem read and writes")
        self.do_rw(0, 0xabcd3210)
        self.done()

    def t_se_cpu(self):
        self.start("check shmem ecc single error (by cpu)")
        sv0 = self.get_supervisor()
        if sv0.cfg != 0:
            self.cmd("shmem-scrubber 0")
        self.cmd("shmem-ecc 1")
        v0 = 0xef012345
        self.do_rw(4, v0)
        self.cmd("shmem-ecc 0")
        sv = self.get_supervisor()
        self.failif(sv.err != sv0.err + 1,
                    "shmem ecc error has not been detected (by cpu)")
        self.failif(sv.sce != sv0.sce,
                    "shmem ecc error has been detected by disabled scrubber")
        v = self.do_rd_cmp(4, v0)
        sv1 = self.get_supervisor()
        self.failif(sv1.err != sv.err,
                    "shmem error must have been fixed")
        self.done()

    def t_se_scrubber(self):
        self.start("check shmem ecc single error (by scrubber)")
        sv0 = self.get_supervisor()
        if sv0.cfg != 1:
            self.cmd("shmem-scrubber 1")
        self.cmd("shmem-ecc 1")
        v0 = 0x99887766
        self.do_rw(12, v0)
        self.cmd("shmem-ecc 0")
        sv = self.get_supervisor()
        self.failif(sv.err != sv0.err,
                    "shmem ecc error has been detected (by cpu)")
        self.failif(sv.sce != sv0.sce + 1,
                    "shmem ecc error has not been detected by the scrubber")
        self.done()

    def t_de_scrubber(self):
        self.start("check shmem ecc double error (scrubber)")
        sv0 = self.get_supervisor()
        if sv0.cfg != 1:
            self.cmd("shmem-scrubber 1")
        self.cmd("shmem-ecc 5")
        v0 = 0x98672354
        self.do_write(8, v0)
        sv = self.get_supervisor()
        self.failif(sv.sde != sv0.sde + 1,
                    "shmem ecc double-error has not been detected")
        self.failif(sv.sce != sv0.sce,
                    "shmem extrac ecc single-error detected")
        self.cmd("shmem-ecc 0")
        self.do_write(8, v0)
        self.done()

    def t_de_cpu(self):
        self.start("check shmem ecc double error (cpu)")
        sv0 = self.get_supervisor()
        self.cmd("shmem-ecc 0x48")
        v0 = 0xb49d54a3
        self.do_write(16, v0)
        self.do_read(16)
        sv = self.get_supervisor()
        self.failif(sv.mde != sv0.mde + 1,
                    "shmem ecc double-error not detected")
        self.cmd("shmem-ecc 0")
        self.do_write(16, v0)
        self.done()

    def run_all_tests(self):
        self.t_rw()
        self.t_se_cpu()
        self.t_se_scrubber()
        self.t_de_scrubber()
        self.t_de_cpu()
        return self.report()

def do_shmem_tests(s):
    shm = shmem_tests(s)
    shm.run_all_tests()


def do_all_tests(s, num):
    mgmt = mgmt_tests(s)
    app0 = app0_tests(s)
    app0.t_load()
    app1 = app1_tests(s)
    app1.load()
    shm = shmem_tests(s)

    for i in range(num):
        print("################ run {} ##############".format(i))
        if not mgmt.run_all_tests():
            break
        if not app0.run_all_tests():
            break
        if not app1.run_all_tests():
            break
        if not shm.run_all_tests():
            break


class MyCSV:
    def __init__(self):
        self.csv = []
        self.prev = {}
        self.change = 'start'

    def add(self, name, val, test=True):
        self.csv.append((name, val))
        if test and not self.change and self.prev[name] != val:
            self.change = name

    def add_percnt(self, name, val, pcnt=5.0):
        self.csv.append((name, val))
        if not self.change:
            val = float(val)
            prev = float(self.prev[name])
            dlt = prev * pcnt / 100.0
            if dlt < 0.0:
                dlt = -dlt
            if val < prev - dlt or val > prev + dlt:
                self.change = name

    def get_header(self):
        return ','.join(t[0] for t in self.csv)

    def get_csv(self):
        return ','.join(str(t[1]) for t in self.csv)

    def get_change(self):
        return self.change

    def has_changed(self):
        return self.change is not None

    def update(self):
        self.prev = {t[0]:t[1] for t in self.csv}
        self.csv = []
        self.change = None

    def discard(self):
        self.csv = []
        self.change = None


def do_plc_supervisor_csv(s, csv, verbose, num):
    p = s.cmd("plc {}".format(num))
    if verbose:
        print("plc{}: {}".format(num, p))
    csv.add('=plc{}='.format(num), '=plc{}='.format(num))
    csv.add('plc{}-reset'.format(num),
            extract(p[0], "plc reset:  ", "  # 1: sys-rst, 2: cpu-rst"))
    csv.add('plc{}-status'.format(num),
            extract(p[1], "plc status: ", "  # 1: fault"))
    p = s.cmd("plc-supervisor {}".format(num))
    if verbose:
        print("plc-supervisor{}: {}".format(num, p))
    csv.add('plc{}-reset-cause'.format(num),
            extract(p[0], "PLC reset cause: ",
                    "  # 1: cpu, 2: iram, 4: dram, 8: wd, 0x10: recov"))
    csv.add('plc{}-iram-se-corr'.format(num),
            extract(p[1], "PLC IRAM corrected errors: "))
    csv.add('plc{}-iram-scrub-cyc'.format(num),
            extract(p[2], "PLC IRAM scrub cycle number: "), False)
    csv.add('plc{}-iram-scrub-se'.format(num),
            extract(p[3], "PLC IRAM corrected errors by scrubber: "))
    csv.add('plc{}-iram-scrub-de'.format(num),
            extract(p[4], "PLC IRAM double errors detected in last cycle: "))
    csv.add('plc{}-dram-se-corr'.format(num),
            extract(p[5], "PLC DRAM corrected errors: "))
    csv.add('plc{}-dram-scrub-cyc'.format(num),
            extract(p[6], "PLC DRAM scrub cycle number: "), False)
    csv.add('plc{}-dram-scrub-se'.format(num),
            extract(p[7], "PLC DRAM corrected errors by scrubber: "))
    csv.add('plc{}-dram-scrub-de'.format(num),
            extract(p[8], "PLC DRAM double errors detected in last cycle: "))
    csv.add('plc{}-scrub-cfg'.format(num),
            extract(p[9], "PLC scrub config: "))
    csv.add('plc{}-cpu-reset'.format(num),
            extract(p[10], "PLC nbr cpu reset: "))
    csv.add('plc{}-user-fault'.format(num),
            extract(p[11], "PLC nbr user fault: "))


def do_charm_tests(s):
    iram_err_addr = 0
    iram_err_cnt = 0
    count = 0
    flag_erase = False

    print("Enable adc+rtp scan")
    s.cmd("i2c-scan adc rtp")

    if flags.flag_load:
        print("load app0")
        s.cmd("plc 0 reset")
        s.cmd("plc 0 iwrite")
        s.cmd("load 0 plc0_charm.elf")
        print("load app1")
        s.cmd("plc 1 reset")
        s.cmd("plc 1 iwrite")
        s.cmd("load 1 plc1_charm.elf")
        s.cmd("read 0xff0")

    if flags.flag_flash:
        flash_id(s)

    print("Self-test for mgmt")
    do_mgmt_tests(s)

    if flags.flag_dry:
        return

    # Find first free page
    if flags.flag_flash:
        pg = flash_find_free(s)
        if pg < 0:
            print("No free page")
            flash_full_erase(s)
            pg = 0
            print("First free page: {}".format(pg))
        off = 0

    mbx1_exp = 0x12345678
    mbx1_cnt = 0
    mbx0_exp = 0x23456789
    mbx0_ncnt = None

    ping_err = 0
    mbx1_err = 0
    mbx0_err = 0

    first = True
    change = True
    prev = {}
    csv = MyCSV()
    while True:
        try:
            t = time.time()
            ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
            ts += ".{}".format(int(t * 10) % 10)  # Append 10th of seconds
            # csv = [('time', ts)]
            if s.trace:
                sys.stderr.write("Date: {}\n".format(ts))

            p = s.cmd("ping")
            if flags.verbose:
                print("Ping: {}".format(p))
            if p[0].find('!') != -1:
                ping_err += 1
            csv.add('ping', ping_err)
            p = s.cmd("fip-err")
            if flags.verbose:
                print("fip-err: {}".format(p))
            csv.add('fip-err', extract(p[0], "fip errors: "))
            csv.add('comm-err', s.nbr_comm_err)

            p = s.cmd("system-read")
            if flags.verbose:
                print("System: {}".format(p))
            csv.add('=i2c=', '=i2c=')
            csv.add('cycle', extract(p[0], "Cycle: "), False)
            csv.add('buf-len', extract(p[1], "Buffer length: "))
            csv.add('buses', extract(p[2], "Buses: ").split(' ')[0])

            # Skip Cycle, Buffer Length, Buses.
            p = p[3:]
            pr = p and p[0].startswith('Tamb')
            if pr == False:
                print(pr)
                print(p)
                break
            csv.add('=adc=', '=adc=')
            csv.add_percnt('Tamb', extract(p[0], "Tamb: ", " C") if pr else 0)
            csv.add_percnt('Tldo', extract(p[1], "Tldo: ", " C") if pr else 0)
            csv.add_percnt('Tfpga', extract(p[2], "Tfpga: ", " C") if pr else 0)
            csv.add_percnt('P12V', extract(p[3], "P12V: ", " V") if pr else 0)
            csv.add_percnt('P12A', extract(p[4], "P12A: ", " A") if pr else 0, 10.0)
            csv.add_percnt('P1.2V', extract(p[5], "P1.2V: ", " V") if pr else 0)
            csv.add_percnt('P3.3V', extract(p[6], "P3.3V: ", " V") if pr else 0)
            csv.add_percnt('Vperi', extract(p[7], "Vperi: ", " V") if pr else 0)
            csv.add_percnt('SBadcT', extract(p[8], "SBadcT: ", " C") if pr else 0)

            if pr:
                p = p[9:]
            csv.add('=fan=', '=fan=')
            pr = p and p[0].startswith('FTP12V')
            csv.add('FTP12V', extract(p[0], "FTP12V: ", " V") if pr else 0)
            csv.add('FTP12A', extract(p[1], "FTP12V: ", " A") if pr else 0)
            csv.add('FTP5V', extract(p[2], "FTP5V: ", " V") if pr else 0)
            csv.add('FTtemp', extract(p[3], "FTtemp: ", " C") if pr else 0)
            csv.add('fan1', extract(p[4], "fan 1: ") if pr else 0)
            csv.add('fan2', extract(p[5], "fan 2: ") if pr else 0)
            csv.add('fan3', extract(p[6], "fan 3: ") if pr else 0)

            if pr:
                p = p[7:]
            csv.add('=rtp=', '=rtp=')
            pr = p and p[0].startswith('RtpP12V')
            csv.add_percnt('RtpP12V', extract(p[0], "RtpP12V: ", " V") if pr else 0)
            csv.add_percnt('RtpP12A', extract(p[1], "RtpP12A: ", " A") if pr else 0)
            csv.add_percnt('RtpP5V', extract(p[2], "RtpP5V: ", " V") if pr else 0)
            csv.add_percnt('RtpP5A', extract(p[3], "RtpP5A: ", " A") if pr else 0)
            csv.add_percnt('RtpadcT', extract(p[4], "RtpadcT: ", " C") if pr else 0)

            p = s.cmd("supervisor")
            if flags.verbose:
                print("supervisor: {}".format(p))
            csv.add('=sup=', '=sup=')
            csv.add('reset-cause', extract(p[0], "reset cause: "))
            csv.add('cpus-reset', extract(p[1], "cpus on reset: "))
            csv.add('iram-se-corr', extract(p[4], "IRAM corrected errors: "))
            csv.add('iram-scrub-cyc',
                    extract(p[5], "IRAM scrub cycle number: "), False)
            csv.add('iram-scrub-se',
                    extract(p[6], "IRAM corrected errors by scrubber: "))
            csv.add('iram-scrub-de',
                    extract(p[8], "IRAM double errors detected in last cycle: "))
            csv.add('dram-se-corr',
                    extract(p[9], "DRAM corrected errors: "))
            csv.add('dram-scrub-cyc',
                    extract(p[10], "DRAM scrub cycle number: "), False)
            csv.add('dram-scrub-se',
                    extract(p[11], "DRAM corrected errors by scrubber: "))
            csv.add('dram-scrub-de',
                    extract(p[13], "DRAM double errors detected in last cycle: "))
            csv.add('insn-bus-err',
                    extract(p[14], "Errors on cpu data bus: "))
            csv.add('data-bus-err',
                    extract(p[15], "Errors on cpu insn bus: "))
            csv.add('ram-scrub-cfg',
                    extract(p[16], "Scrubber config: "))
            csv.add('iram-ecc-addr',
                    extract(p[19], "IRAM ecc addr: "))
            csv.add('=sta=', '=sta=')

            p = s.cmd("stat-regs")
            if flags.verbose:
                print("stat-regs: {}".format(p))
            csv.add('adc-reset',
                    extract(p[0], "nbr adc reset: "))
            csv.add('mm-fantray-reset',
                    extract(p[1], "nbr fantray reset: "))
            csv.add('ratopus-errs',
                    extract(p[2], "nbr ratopus errors: "))
            csv.add('min-watchdog',
                    extract(p[3], "minimum watchdog count: "))
            csv.add('nbr-recovery',
                    extract(p[4], "nbr recovery: "))
            csv.add('nbr-reset',
                    extract(p[5], "nbr resets: "))
            csv.add('last-reset',
                    extract(p[6], "last reset: "))
            csv.add('cycle-count',
                    extract(p[7], "cycle count: "), False),
            csv.add('fip-cmds',
                    extract(p[8], "fip commands: "), False)
            csv.add('fip-pad',
                    extract(p[9], "fip pad: "), False)
            if False and len(p) > 12:
                csv.add('fip-writes',
                        extract(p[12], "fip writes: "))
                csv.add('wr-addr',
                        extract(p[13], "write addr: "))

            #### PLC-0 (supervisor and shm)
            do_plc_supervisor_csv(s, csv, flags.verbose, 0)
            if True:
                csv.add('=shm0=', '=shm0=')
                p = s.cmd("last-data")
                if flags.verbose:
                    print("last-data: {}".format(p))
                mbx0_cnt = int(extract(p[0], "count: "))
                if mbx0_ncnt is not None:
                    mbx0_dat = extract(p[1], "data: ")
                    mbx0_dat = mbx0_dat.split()
                    if mbx0.len() > 7:
                        mbx0_val = int(mbx0_dat[4], 16)
                        if mbx0_exp != mbx0_val:
                            print("incorrect received data")
                            mbx0_err += 1
                    else:
                        print("invalid last-data")

                mbx0_nval = rnd.randint(0, (1<<8) - 1)
                s.cmd("data 0 0 0 0 0x{:02x} 0 0 0".format(mbx0_nval))
                mbx0_exp = (mbx0_nval + 1) & 0xff
                csv.add('shm0-err', mbx0_err)
                csv.add('shm0-cnt', mbx0_cnt, False)

            shm = do_shmem_supervisor(s)
            csv.add('=shm=', '=shm=')
            csv.add('shm-se', shm.err)
            csv.add('shm-de-mgmt', shm.mde)
            csv.add('shm-de-app', shm.ade)
            csv.add('shm-src-cyc', shm.cyc, False)
            csv.add('shm-src-se', shm.sce)
            csv.add('shm-scr-de', shm.sde)
            csv.add('shm-scr-cfg', shm.cfg)
            csv.add('shm-scr-per', shm.per)

            #### PLC-1 (supervisor and mbox)
            do_plc_supervisor_csv(s, csv, flags.verbose, 1)
            csv.add('=mbx1=', '=mbx1=')
            mbx1_sta, mbx1_val = do_extract_mbox1(s)
            if mbx1_sta != 1:
                print("mailbox 1 not ready")
                mbx1_err += 1
            else:
                if mbx1_val != mbx1_exp:
                    print("mailbox 1 error: expect {:08x}, read {:08x}".format(
                        mbx1_exp, mbx1_val))
                    mbx1_err += 1
                    mbx1_cnt = (mbx1_val >> 16) & 0xffff
            # Always write so we are not stuck
            mbx1_nval = rnd.randint(0, (1<<32) - 1)
            s.cmd("mbox 1 0x{:08x}".format(mbx1_nval))
            mbx1_cnt = (mbx1_cnt + 1) & 0xffff
            mbx1_exp = ((~mbx1_nval) & 0x0000ffff) | (mbx1_cnt << 16)
            #mbx1_exp = (mbx1_nval + 1) & 0xffffffff
            csv.add('mbx1-err', mbx1_err)
            csv.add('mbx1-cnt', mbx1_cnt, False)

            csv.add('flash-err', s.nbr_flash_err)

            if first:
                if flags.log:
                    log.write('time,' + csv.get_header() + '\n')
                print('time,' + csv.get_header())
                first = False

            line = ts + ',' + csv.get_csv()
            if flags.log:
                flags.log.write(line + '\n')
                flags.log.flush()
            if csv.has_changed():
                print(line + '  ## ' + csv.get_change())
                sys.stdout.flush()
            elif flags.flag_all or (flags.flag_monitor and (count % 100 == 0)):
                print(line)
                sys.stdout.flush()
            csv.update()
            count += 1

            if flags.flag_flash:
                if flag_erase:
                    flash_page_erase(s, pg)
                flash_prg(s, pg, off)
                pg = (pg + 1) & 0xff
                if pg == 0:
                    off = (off + 32) & 0xff
                    flag_erase = (off == 0)

            if flags.flag_iram_err:
                iram_err_cnt = (iram_err_cnt + 1) & 3
                if iram_err_cnt == 0:
                    print("iram err at 0x{:x}".format(iram_err_addr))
                    s.cmd("write 0x80050 0x{:x}".format(iram_err_addr | 1))
                    iram_err_addr += 4
                    if iram_err_addr > 0xa00:
                        iram_err_addr = 0


        except CommError:
            print("# Comm error")
            csv.discard()


def run():
    s = FipUrvDrv(trace=False)
    # s.cmd("slots")

    action = None

    for arg in sys.argv[1:]:
        if arg[0] == '-':
            if arg == '-v':
                flags.verbose = True
            elif arg == '--dry' or arg == '-n':
                flags.flag_dry = True
            elif arg == '--no-flash':
                flags.flag_flash = False
            elif arg == '--iram-err':
                flags.flag_iram_err = True
            elif arg == '-t':
                s.trace = True
            elif arg == '--no-load':
                flags.flag_load = False
            elif arg == '--all':
                # Print all the CSV (otherwise, only on a change)
                flags.flag_all = True
            elif arg == '--monitor':
                # 1/200 CSV
                flags.flag_monitor = True
            elif arg[:2] == '-l':
                flags.log = open(arg[2:], "a")
            else:
                print("unknown option: {}".format(arg))
                sys.exit(1)
        else:
            action = arg

    if action is None or action == 'charm':
        do_charm_tests(s)
    elif action == 'mgmt':
        do_mgmt_tests(s)
    elif action == 'app0' or action == 'plc0':
        do_app0_tests(s)
    elif action == 'app1':
        do_app1_tests(s)
    elif action == 'shmem':
        do_shmem_tests(s)
    elif action == 'all':
        do_all_tests(s, 1)
    elif action == 'repeat':
        do_all_tests(s, 1000)
    elif action == 'reset':
        s.cmd("fip-reset")
    else:
        print("unknown action: {}".format(action))

    s.close()

run()
